# ----------------------------------------------------------------------
# |
# |  DoneManager.py
# |
# |  David Brownell <db@DavidBrownell.com>
# |      2023-12-19 10:50:59
# |
# ----------------------------------------------------------------------
# |
# |  Copyright David Brownell 2023-24
# |  Distributed under the MIT License.
# |
# ----------------------------------------------------------------------
"""Contains functionality used with DoneManager objects"""

import datetime
import itertools
import os
import sys
import time
import traceback

from contextlib import contextmanager
from dataclasses import dataclass, field, InitVar
from enum import auto, Flag
from types import TracebackType
from typing import Any, Callable, Iterator, Optional, Protocol, Type as PythonType

from dbrownell_Common.ContextlibEx import ExitStack
from dbrownell_Common import TextwrapEx
from dbrownell_Common.Streams.Capabilities import Capabilities
from dbrownell_Common.Streams.StreamDecorator import StreamDecorator, TextWriterT


# ----------------------------------------------------------------------
DISPLAYED_EXCEPTION_ATTRIBUTE_NAME: str = "_done_manager_displayed_exception___"


# ----------------------------------------------------------------------
# |  Optional Functionality if typer is installed
try:
    _original_exception_hook = None  # pylint: disable=invalid-name

    # ----------------------------------------------------------------------
    def _ExceptionHook(
        exc_type: PythonType[BaseException],
        exc_value: BaseException,
        tb: Optional[TracebackType],
    ) -> Any:
        if getattr(exc_value, DISPLAYED_EXCEPTION_ATTRIBUTE_NAME, False):
            return

        assert _original_exception_hook is not None
        _original_exception_hook(exc_type, exc_value, tb)

    # ----------------------------------------------------------------------

    # This environment name is defined by typer. Setting it signals that typer
    # should not inject its own exception hook.
    os.environ["_TYPER_STANDARD_TRACEBACK"] = "1"

    # Monkey-patch typer's _original_except_hook if typer has already been imported
    if "typer" in sys.modules:
        typer = sys.modules["typer"]

        _original_exception_hook = (
            typer.main._original_except_hook  # pylint: disable=protected-access
        )  # pylint: disable=protected-access
        typer.main._original_except_hook = _ExceptionHook  # pylint: disable=protected-access
    else:
        _original_exception_hook = sys.excepthook
        sys.excepthook = _ExceptionHook

        import typer

    from click.exceptions import ClickException

    # ----------------------------------------------------------------------
    def ExitWithTyper(
        result: int,
    ) -> None:
        raise typer.Exit(result)

    # ----------------------------------------------------------------------
    def ShouldRaiseExceptionWithTyper(
        exception: Exception,
    ) -> bool:
        return isinstance(exception, (typer.Exit, typer.Abort, ClickException))

    # ----------------------------------------------------------------------

    Exit = ExitWithTyper
    ShouldRaiseException = ShouldRaiseExceptionWithTyper

except ImportError:
    # ----------------------------------------------------------------------
    class ExitException(Exception):
        """Exception generated in place of a call to sys.exit."""

        # ----------------------------------------------------------------------
        def __init__(
            self,
            result: int,
        ):
            self.result = result

            super(ExitException, self).__init__(
                "The process is returning with a result of '{}'.".format(result)
            )

    # ----------------------------------------------------------------------
    def ExitNoTyper(
        result: int,
    ):
        raise ExitException(result)

    # ----------------------------------------------------------------------
    def ShouldRaiseExceptionNoTyper(
        exception: Exception,  # pylint: disable=unused-argument
    ) -> bool:
        return False

    # ----------------------------------------------------------------------

    Exit = ExitNoTyper
    ShouldRaiseException = ShouldRaiseExceptionNoTyper


# ----------------------------------------------------------------------
# |  Optional functionality if rich is installed
try:
    import rich

    # ----------------------------------------------------------------------
    def ShowCursorWithRich(
        show: bool,
    ) -> None:
        rich.get_console().show_cursor(show)

    # ----------------------------------------------------------------------

    ShowCursor = ShowCursorWithRich

except ImportError:
    # ----------------------------------------------------------------------
    def ShowCursorNoRich(
        show: bool,  # pylint: disable=unused-argument
    ) -> None:
        pass

    # ----------------------------------------------------------------------

    ShowCursor = ShowCursorNoRich


# ----------------------------------------------------------------------
class DoneManagerException(Exception):
    """Exception whose call stack is not displayed when caught within an active DoneManager."""

    pass  # pylint: disable=unnecessary-pass


# ----------------------------------------------------------------------
class Flags(Flag):
    """Flags used to filter the output for a DoneManager"""

    VerboseFlag = auto()
    DebugFlag = auto()

    # Amalgamations
    Standard = 0
    Verbose = VerboseFlag
    Debug = VerboseFlag | DebugFlag

    # ----------------------------------------------------------------------
    @classmethod
    def Create(
        cls,
        *,
        verbose: bool = False,
        debug: bool = False,
    ) -> "Flags":
        if debug:
            flag = cls.Debug
        elif verbose:
            flag = cls.Verbose
        else:
            flag = cls.Standard

        return flag


# ----------------------------------------------------------------------
@dataclass(frozen=True)
class Args:
    """Arguments provided when creating DoneManager instances"""

    # ----------------------------------------------------------------------
    # |
    # |  Public Data
    # |
    # ----------------------------------------------------------------------
    heading_param: InitVar[str]
    heading: str = field(init=False)

    done_suffix_or_suffixes: InitVar[
        None | Callable[[], Optional[str]] | list[Callable[[], Optional[str]]]
    ] = None
    done_suffixes: list[Callable[[], Optional[str]]] = field(init=False)

    prefix: None | str | Callable[[], Optional[str]] = field(kw_only=True, default=None)
    suffix: None | str | Callable[[], Optional[str]] = field(kw_only=True, default=None)

    line_prefix: Optional[StreamDecorator.PrefixOrSuffixT] = field(kw_only=True, default="  ")

    display: bool = field(kw_only=True, default=True)  # Display done information
    display_result: bool = field(kw_only=True, default=True)  # Display the result
    display_time: bool = field(kw_only=True, default=True)  # Display the time delta

    display_exceptions: bool = field(
        kw_only=True, default=True
    )  # Display exceptions when they are encountered; exceptions will percolate if not displayed.
    display_exception_details: bool = field(
        kw_only=True, default=False
    )  # Do not display exception details (such as the call stack)
    suppress_exceptions: bool = field(kw_only=False, default=False)  # Do not let exceptions propagate

    preserve_status: bool = field(kw_only=True, default=True)

    # ----------------------------------------------------------------------
    # |
    # |  Public Methods
    # |
    # ----------------------------------------------------------------------
    def __post_init__(
        self,
        heading_param: str,
        done_suffix_or_suffixes: (None | Callable[[], Optional[str]] | list[Callable[[], Optional[str]]]),
    ) -> None:
        if done_suffix_or_suffixes is None:
            suffixes = []
        elif isinstance(done_suffix_or_suffixes, list):
            suffixes = done_suffix_or_suffixes
        else:
            suffixes = [
                done_suffix_or_suffixes,
            ]

        object.__setattr__(self, "done_suffixes", suffixes)

        if self.suppress_exceptions and not self.display_exceptions:
            raise Exception("Exceptions should not be suppressed when they are not displayed.")

        if not self.display:
            object.__setattr__(self, "line_prefix", None)

        heading = "{}{}".format(
            heading_param.rstrip("\n").rstrip("..."),
            "..." if self.display and heading_param and self.line_prefix else "",
        )

        object.__setattr__(self, "heading", heading)


# ----------------------------------------------------------------------
@dataclass(frozen=True)
class TopLevelArgs(Args):
    """Arguments when creating top-level/non-nested DoneManager instances"""

    # ----------------------------------------------------------------------
    flags: Flags = field(kw_only=True, default=Flags.Standard)
    num_cols: int = field(kw_only=True, default=Capabilities.DEFAULT_COLUMNS)


# ----------------------------------------------------------------------
@dataclass
class DoneManager:
    """Object that helps when writing nested output."""

    # ----------------------------------------------------------------------
    result: int = field(init=False, default=0)

    _stream: StreamDecorator

    flags: Flags = field(kw_only=True)
    heading: str = field(kw_only=True)
    preserve_status: bool = field(kw_only=True)

    num_cols: int = field(kw_only=True)

    _line_prefix: str = field(init=False)
    _status_line_prefix: str = field(init=False)
    _num_status_cols: int = field(init=False)

    _wrote_content: bool = field(init=False, default=False)
    _wrote_status: bool = field(init=False, default=False)
    _prev_status_content: list[str] = field(init=False, default_factory=list)

    # ----------------------------------------------------------------------
    @classmethod
    @contextmanager
    def Create(
        cls,
        stream: TextWriterT,
        *top_level_args,  # See `TopLevelArgs`
        **top_level_kwargs,  # See `TopLevelArgs`
    ) -> Iterator["DoneManager"]:
        args = TopLevelArgs(*top_level_args, **top_level_kwargs)

        with cls._CreateImpl(
            stream,
            args,
            flags=args.flags,
            num_cols=args.num_cols,
        ) as dm:
            try:
                yield dm
            finally:
                ShowCursor(True)

    # ----------------------------------------------------------------------
    @classmethod
    @contextmanager
    def CreateCommandLine(
        cls,
        stream: TextWriterT = sys.stdout,
        *,
        flags: Flags = Flags.Standard,
        display: bool = True,
        **kwargs,
    ) -> Iterator["DoneManager"]:
        """Creates a DoneManager instance suitable for use with command-line functionality."""

        if display:
            prefix = "\nResults: "
            suffix = "\n"
        else:
            prefix = ""
            suffix = ""

        with cls.Create(
            stream,
            "\n",
            line_prefix="",
            prefix=prefix,
            suffix=suffix,
            flags=flags,
            display=display,
            **kwargs,
        ) as dm:
            is_exceptional = False

            try:
                yield dm

            except:
                is_exceptional = True
                raise

            finally:
                if not is_exceptional:
                    Exit(dm.result)

    # ----------------------------------------------------------------------
    @property
    def is_verbose(self) -> bool:
        return bool(self.flags & Flags.VerboseFlag)

    @property
    def is_debug(self) -> bool:
        return bool(self.flags & Flags.DebugFlag)

    @property
    def capabilities(self) -> Capabilities:
        return Capabilities.Get(self._stream)

    # ----------------------------------------------------------------------
    def isatty(self) -> bool:
        return self._stream.isatty()

    # ----------------------------------------------------------------------
    def ExitOnError(self) -> None:
        """Exits if the result is < 0"""

        if self.result < 0:
            Exit(self.result)

    # ----------------------------------------------------------------------
    def ExitOnWarning(self) -> None:
        """Exits if the result is > 0"""

        if self.result > 0:
            Exit(self.result)

    # ----------------------------------------------------------------------
    def WriteLine(
        self,
        content: str,
    ) -> None:
        """Writes a line without decoration; status information is preserved or cleared based on the `preserve_status` flag."""

        # ----------------------------------------------------------------------
        def CreateStandardText(
            value: str,
            *,
            supports_colors: bool = True,  # pylint: disable=unused-argument
            decorate_every_line: bool = False,  # pylint: disable=unused-argument
        ) -> str:
            return value

        # ----------------------------------------------------------------------

        self._WriteImpl(CreateStandardText, content)

    # ----------------------------------------------------------------------
    def WriteSuccess(
        self,
        content: str,
    ) -> None:
        """Writes a line decorated with a success prefix; status information is preserved or cleared based on the `preserve_status` flag."""

        self._WriteImpl(TextwrapEx.CreateSuccessText, content)

    # ----------------------------------------------------------------------
    def WriteError(
        self,
        content: str,
        *,
        update_result: bool = True,
    ) -> None:
        """Writes a line decorated with an error prefix; status information is preserved or cleared based on the `preserve_status` flag."""

        self._WriteImpl(TextwrapEx.CreateErrorText, content)

        if update_result and self.result >= 0:
            self.result = -1

    # ----------------------------------------------------------------------
    def WriteWarning(
        self,
        content: str,
        *,
        update_result: bool = True,
    ) -> None:
        """Writes a line decorated with a warning prefix; status information is preserved or cleared based on the `preserve_status` flag."""

        self._WriteImpl(TextwrapEx.CreateWarningText, content)

        if update_result and self.result == 0:
            self.result = 1

    # ----------------------------------------------------------------------
    def WriteInfo(
        self,
        content: str,
    ) -> None:
        """Writes a line decorated with an info prefix; status information is preserved or cleared based on the `preserve_status` flag."""

        self._WriteImpl(TextwrapEx.CreateInfoText, content)

    # ----------------------------------------------------------------------
    def WriteVerbose(
        self,
        content: str,
    ) -> None:
        """\
        Writes verbose content if the verbose flag is set. Use this functionality when you want
        only the first line to include the verbose decorator. Use `YieldVerboseStream` when you
        want every line to include the verbose decorator.

        Status information is preserved or cleared based on the `preserve_status` flag.
        """

        if self.is_verbose:
            self._WriteImpl(TextwrapEx.CreateVerboseText, content)

    # ----------------------------------------------------------------------
    def WriteDebug(
        self,
        content: str,
    ) -> None:
        """\
        Writes debug content if the debug flag is set. Use this functionality when you want
        only the first line to include the debug decorator. Use `YieldDebugStream` when you
        want every line to include the debug decorator.

        Status information is preserved or cleared based on the `preserve_status` flag.
        """

        if self.is_debug:
            self._WriteImpl(TextwrapEx.CreateDebugText, content)

    # ----------------------------------------------------------------------
    def WriteStatus(
        self,
        content: str,
    ) -> None:
        """\
        Writes status information; status information is temporal, and is replaced each time new
        status is added.
        """

        self._WriteStatus(content)

    # ----------------------------------------------------------------------
    def ClearStatus(self) -> None:
        """Clears any status information currently displayed"""

        if self._prev_status_content:
            self._WriteStatus("", update_prev_status=False)
            self._prev_status_content = []

    # ----------------------------------------------------------------------
    def PreserveStatus(self) -> None:
        """Persists any status information currently displayed so that it will not be overwritten"""

        if self._prev_status_content:
            # Move the cursor back to where it would be after writing the status messages normally.
            self._stream.write("\033[{}B".format(len(self._prev_status_content)))
            self._stream.write("\r")

            self._prev_status_content = []

    # ----------------------------------------------------------------------
    @contextmanager
    def Nested(
        self,
        *args,  # See `Args`
        **kwargs,  # See `Args`
    ) -> Iterator["DoneManager"]:
        """Creates a nested DoneManager"""

        with self.YieldStream() as stream:
            with self._CreateNestedImpl(stream, *args, **kwargs) as dm:
                yield dm

    # ----------------------------------------------------------------------
    @contextmanager
    def VerboseNested(
        self,
        *args,  # See `Args`
        **kwargs,  # See `Args`
    ) -> Iterator["DoneManager"]:
        """Creates a nested DoneManager if the verbose flag is set"""

        with self.YieldVerboseStream() as verbose_stream:
            with self._CreateNestedImpl(verbose_stream, *args, **kwargs) as dm:
                yield dm

    # ----------------------------------------------------------------------
    @contextmanager
    def DebugNested(
        self,
        *args,  # See `Args`
        **kwargs,  # See `Args`
    ) -> Iterator["DoneManager"]:
        """Creates a nested DoneManager if the debug flag is set"""

        with self.YieldDebugStream() as debug_stream:
            with self._CreateNestedImpl(debug_stream, *args, **kwargs) as dm:
                yield dm

    # ----------------------------------------------------------------------
    @contextmanager
    def YieldStream(self) -> Iterator[StreamDecorator]:
        """Provides scoped access to the underlying stream; writing to this stream will include line prefixes"""

        if self.preserve_status:
            self.PreserveStatus()
        else:
            self.ClearStatus()

        yield self._stream
        self._stream.flush()

    # ----------------------------------------------------------------------
    @contextmanager
    def YieldVerboseStream(self) -> Iterator[StreamDecorator]:
        """\
        Provides scoped access to a verbose-version of the underlying stream; writing to this
        stream will include verbose line prefixes. Use this functionality when you want all
        lines to have the verbose prefix decorator. Use `WriteVerbose` when you only want the
        first line to include the decorator.
        """

        if self.preserve_status:
            self.PreserveStatus()
        else:
            self.ClearStatus()

        if self.is_verbose:
            stream = StreamDecorator(
                self._stream,
                line_prefix=TextwrapEx.CreateVerbosePrefix(Capabilities.Get(self._stream).supports_colors),
                decorate_empty_lines=True,
            )
        else:
            stream = StreamDecorator(None)

        yield stream

        if stream.wrote_content:
            self._wrote_content = True
            stream.flush()

    # ----------------------------------------------------------------------
    @contextmanager
    def YieldDebugStream(self) -> Iterator[StreamDecorator]:
        """\
        Provides scoped access to a debug-version of the underlying stream; writing to this
        stream will include debug line prefixes. Use this functionality when you want all
        lines to have the debug prefix decorator. Use `WriteDebug` when you only want the
        first line to include the decorator.
        """

        if self.preserve_status:
            self.PreserveStatus()
        else:
            self.ClearStatus()

        if self.is_debug:
            stream = StreamDecorator(
                self._stream,
                line_prefix=TextwrapEx.CreateDebugPrefix(Capabilities.Get(self._stream).supports_colors),
                decorate_empty_lines=True,
            )
        else:
            stream = StreamDecorator(None)

        yield stream

        if stream.wrote_content:
            self._wrote_content = True
            stream.flush()

    # ----------------------------------------------------------------------
    @contextmanager
    def YieldStdout(self) -> Iterator[StreamDecorator.YieldStdoutContext]:
        """\
        Provides scoped access to `sys.stdout` (if possible); writing to this stream will NOT include line prefixes.

        Note that the cursor should be restored to its original location (when yielded by
        this generator) when control is restored to this generator and the `persist_content`
        is set to False.
        """

        if self.preserve_status:
            self.PreserveStatus()
        else:
            self.ClearStatus()

        is_interactive = self.capabilities.is_interactive

        with self._stream.YieldStdout() as context:
            try:
                if not self._wrote_content:
                    context.stream.write("\n")

                yield context

            finally:
                # We need to get the cursor back to where it was before we
                # yielded the stream. This logic depends upon if anything
                # had been written using this DoneManager before the stream
                # was yielded and if the caller wants to preserve anything
                # that was written.
                if context.stream is sys.stdout:
                    if is_interactive:
                        sys.stdout.write("\r")

                    if context.persist_content:
                        if self._wrote_content:
                            sys.stdout.write("\n")

                        sys.stdout.write(self._line_prefix)

                    elif is_interactive and not context.persist_content and not self._wrote_content:
                        # Move up a line, write the whitespace and heading
                        sys.stdout.write(
                            "\033[1A{}{}".format(self._line_prefix, self.heading),
                        )
                        sys.stdout.flush()

    # ----------------------------------------------------------------------
    def __post_init__(self):
        self._line_prefix = self._stream.GetCompleteLinePrefix(include_self=False)
        self._status_line_prefix = self._line_prefix + self._stream.GetLinePrefix(len(self._line_prefix))

        complete_line_prefix = self._stream.GetCompleteLinePrefix(include_self=True)
        len_complete_line_prefix = len(complete_line_prefix)

        assert self.num_cols > len_complete_line_prefix, (self.num_cols, len_complete_line_prefix)
        self._num_status_cols = self.num_cols - len_complete_line_prefix

    # ----------------------------------------------------------------------
    # |
    # |  Private Types
    # |
    # ----------------------------------------------------------------------
    class _CreateDecoratedTextT(Protocol):
        def __call__(
            self,
            value: str,
            *,
            supports_colors: bool = True,
            decorate_every_line: bool = False,
        ) -> str: ...  # pragma: no cover

    # ----------------------------------------------------------------------
    # |
    # |  Private Methods
    # |
    # ----------------------------------------------------------------------
    @classmethod
    @contextmanager
    def _CreateImpl(
        cls,
        stream: TextWriterT,
        args: Args,
        *,
        flags: Flags,
        num_cols: int,
    ) -> Iterator["DoneManager"]:
        if args.heading:
            stream.write(args.heading)
            stream.flush()

        instance = cls(
            StreamDecorator(
                stream,
                prefix="\n" if args.heading else "",
                line_prefix=args.line_prefix,
                decorate_empty_lines=False,
            ),
            flags=flags,
            heading=args.heading,
            preserve_status=args.preserve_status,
            num_cols=num_cols,
        )

        capabilities = Capabilities.Get(stream)

        if capabilities.supports_colors:
            success_color_on = TextwrapEx.SUCCESS_COLOR_ON
            error_color_on = TextwrapEx.ERROR_COLOR_ON
            warning_color_on = TextwrapEx.WARNING_COLOR_ON
            color_off = TextwrapEx.COLOR_OFF
        else:
            success_color_on = ""
            error_color_on = ""
            warning_color_on = ""
            color_off = ""

        time_delta: Optional[str] = None
        done_suffixes: list[Callable[[], Optional[str]]] = []

        if args.display_result:
            if capabilities.supports_colors:
                # ----------------------------------------------------------------------
                def DisplayResult() -> str:
                    if instance.result < 0:
                        color_on = error_color_on
                    elif instance.result > 0:
                        color_on = warning_color_on
                    else:
                        color_on = success_color_on

                    return "{}{}{}".format(color_on, instance.result, color_off)

                # ----------------------------------------------------------------------

                display_result_func = DisplayResult
            else:
                display_result_func = lambda: str(instance.result)

            done_suffixes.append(display_result_func)

        if args.display_time:
            done_suffixes.append(lambda: str(time_delta) if time_delta is not None else "<Unknown>")

        done_suffixes += args.done_suffixes

        start_time = time.perf_counter()

        # ----------------------------------------------------------------------
        def GetStringValue(
            value: None | str | Callable[[], Optional[str]],
        ) -> Optional[str]:
            if value is None:
                return None

            if isinstance(value, str):
                return value

            if callable(value):
                return value()

            assert False, value  # pragma: no cover

        # ----------------------------------------------------------------------
        def OnExit():
            instance._OnExit()  # pylint: disable=protected-access

            # Display the prefix
            prefix_value = GetStringValue(args.prefix)
            if prefix_value is not None:
                stream.write(prefix_value)

            # Display the content
            if args.display:
                suffixes: list[str] = []

                for done_suffix in done_suffixes:
                    result = done_suffix()
                    if result is not None:
                        suffixes.append(result)

                if suffixes:
                    content = "DONE! ({})\n".format(", ".join(suffixes))
                else:
                    content = "DONE!\n"

                stream.write(content)

            # Display the suffix
            suffix_value = GetStringValue(args.suffix)
            if suffix_value is not None:
                stream.write(suffix_value)

        # ----------------------------------------------------------------------

        with ExitStack(OnExit):
            try:
                yield instance

            except Exception as ex:
                if ShouldRaiseException(ex):
                    raise

                if instance.result >= 0:
                    instance.result = -1

                if (
                    args.display_exceptions
                    # Do not display an exception that has already been displayed
                    and not getattr(ex, DISPLAYED_EXCEPTION_ATTRIBUTE_NAME, False)
                ):
                    object.__setattr__(ex, DISPLAYED_EXCEPTION_ATTRIBUTE_NAME, True)

                    if not isinstance(ex, DoneManagerException) and (
                        flags & Flags.Debug or args.display_exception_details
                    ):
                        exception_content = traceback.format_exc()
                    else:
                        exception_content = str(ex)

                    if instance._stream.col_offset != 0:
                        instance._stream.write("\n")

                    instance._stream.write(
                        TextwrapEx.CreateErrorText(
                            exception_content,
                            supports_colors=capabilities.supports_colors,
                        ),
                    )

                    instance._stream.write("\n")

                if not args.suppress_exceptions:
                    raise

            finally:
                current_time = time.perf_counter()

                assert start_time <= current_time, (start_time, current_time)

                time_delta = str(datetime.timedelta(seconds=current_time - start_time))

    # ----------------------------------------------------------------------
    @contextmanager
    def _CreateNestedImpl(
        self,
        stream: StreamDecorator,
        *args,
        **kwargs,
    ) -> Iterator["DoneManager"]:
        if "preserve_status" not in kwargs:
            kwargs["preserve_status"] = False

        with self.__class__._CreateImpl(  # pylint: disable=protected-access
            stream,
            Args(*args, **kwargs),
            flags=self.flags,
            num_cols=self.num_cols,
        ) as dm:
            try:
                yield dm
            finally:
                if (dm.result < 0 and self.result >= 0) or (dm.result > 0 and self.result == 0):
                    self.result = dm.result

    # ----------------------------------------------------------------------
    def _WriteImpl(
        self,
        create_text_func: "DoneManager._CreateDecoratedTextT",
        content: str,
    ) -> None:
        if self._prev_status_content:
            self._WriteStatus("", update_prev_status=False)

        content = create_text_func(
            content,
            supports_colors=Capabilities.Get(self._stream).supports_colors,
        )

        if not content.endswith("\n"):
            content += "\n"

        self._stream.write(content)
        self._wrote_content = True

        if self._prev_status_content:
            self._WriteStatus(self._prev_status_content, update_prev_status=False)

    # ----------------------------------------------------------------------
    def _WriteStatus(
        self,
        content: str | list[str],
        *,
        update_prev_status: bool = True,
    ) -> None:
        if not self.capabilities.is_interactive:
            return

        ShowCursor(not bool(content))

        # Prepare the content
        lines: list[str] = []
        blank_lines: list[str] = []

        if isinstance(content, list):
            lines = content
            assert not update_prev_status
        else:
            content_lines = content.split("\n")

            if not content_lines[-1]:
                del content_lines[-1]

            for line in content_lines:
                line = "{}{}".format(self._status_line_prefix, line)
                line = TextwrapEx.BoundedLJust(line, self._num_status_cols)

                lines.append("\r{}\n".format(line))

            if len(self._prev_status_content) > len(lines):
                blank_lines += [
                    "\r{}\n".format("".ljust(self._num_status_cols)),
                ] * (len(self._prev_status_content) - len(lines))

        # Write the content
        for line in itertools.chain(lines, blank_lines):
            self._stream.write(line)

        # Move the cursor up to the position that it would be in to write a standard message
        self._stream.write("\033[{}A\r".format(len(lines) + len(blank_lines)))

        self._wrote_status = True

        if update_prev_status:
            self._prev_status_content = lines

    # ----------------------------------------------------------------------
    def _OnExit(self) -> None:
        if self.preserve_status:
            self.PreserveStatus()
        else:
            self.ClearStatus()

            if self.capabilities.is_interactive and not self._wrote_content and self._wrote_status:
                # Move up a line and recreate the heading
                self._stream.write("\033[1A\r{}{}".format(self._line_prefix, self.heading))
                self._stream.flush()



# ----------------------------------------------------------------------
# |
# |  SubprocessEx_UnitTest.py
# |
# |  David Brownell <db@DavidBrownell.com>
# |      2023-12-25 13:17:39
# |
# ----------------------------------------------------------------------
# |
# |  Copyright David Brownell 2023-24
# |  Distributed under the MIT License.
# |
# ----------------------------------------------------------------------
"""Unit tests for SubprocessEx.py."""

import re
import textwrap

import pytest
from unittest.mock import MagicMock as Mock

from dbrownell_Common.SubprocessEx import *
from dbrownell_Common.Streams.Capabilities import Capabilities


# ----------------------------------------------------------------------
class TestRun:
    # ----------------------------------------------------------------------
    def test_Standard(self):
        result = Run("echo Hello World!")

        assert result.returncode == 0
        assert result.output == "Hello World!\n"

        try:
            result.RaiseOnError()
        except Exception:
            assert False, "An exception should not have been raised"

    # ----------------------------------------------------------------------
    def test_Error(self):
        result = Run("this_command_does_not_exist")

        assert result.returncode != 0
        assert result.output != ""

        with pytest.raises(
            Exception,
            match=re.compile(
                textwrap.dedent(
                    """\
                    Command Line
                    ------------
                    this_command_does_not_exist

                    Output
                    ------
                    .+
                    """,
                ),
                re.DOTALL | re.MULTILINE,
            ),
        ):
            result.RaiseOnError()

    # ----------------------------------------------------------------------
    class TestColors:
        command_line = '''python -c "import sys; from dbrownell_Common.TextwrapEx import CreateErrorText; from dbrownell_Common.Streams.Capabilities import Capabilities; print(CreateErrorText('Hello!', supports_colors=Capabilities.Get(sys.stdout).supports_colors));"'''

        # ----------------------------------------------------------------------
        def test_ColorsTrue(self):
            result = Run(self.command_line, supports_colors=True)

            assert result.returncode == 0
            assert result.output == "\x1b[31;1mERROR:\x1b[0m Hello!\n"

        # ----------------------------------------------------------------------
        def test_ColorsFalse(self):
            result = Run(self.command_line, supports_colors=False)

            assert result.returncode == 0
            assert result.output == "ERROR: Hello!\n"

    # ----------------------------------------------------------------------
    def test_Unicode(self):
        result = Run('''python -c "print('🔥')"''')

        assert result.returncode == 0
        assert result.output == "🔥\n"

    # ----------------------------------------------------------------------
    def test_Env(self):
        if os.name.lower() == "nt":
            command_line = "echo _%MY_TEST_VAR%_"
            expected_not_set_output = "_%MY_TEST_VAR%_\n"
        else:
            command_line = "echo _${MY_TEST_VAR}_"
            expected_not_set_output = "__\n"

        # Var not set
        result = Run(command_line)

        assert result.returncode == 0
        assert result.output == expected_not_set_output

        # Var set
        result = Run(
            command_line,
            env={
                "MY_TEST_VAR": "The value",
            },
        )

        # assert result.returncode == 0
        assert result.output == "_The value_\n"


# ----------------------------------------------------------------------
class TestStream:
    # ----------------------------------------------------------------------
    def test_Standard(self):
        mock = Mock()

        result = Stream("echo Hello World", mock)

        assert result == 0
        assert mock.write.call_count == len("Hello World\n")

    # ----------------------------------------------------------------------
    def test_LineDelimited(self):
        mock = Mock()

        result = Stream("echo Hello World", mock, line_delimited_output=True)

        assert result == 0

        assert mock.write.call_count == 2
        assert mock.write.call_args_list[0].args[0] == "Hello World\n"
        assert mock.write.call_args_list[1].args[0] == "\n"

        assert mock.flush.call_count == 1

    # ----------------------------------------------------------------------
    def test_LineDelimitedPartialContent(self):
        mock = Mock()

        result = Stream(
            '''python -c "import sys; sys.stdout.write('Hello\\nWorld')"''',
            mock,
            line_delimited_output=True,
        )

        assert result == 0

        assert mock.write.call_count == 2
        assert mock.write.call_args_list[0].args[0] == "Hello\n"
        assert mock.write.call_args_list[1].args[0] == "World\n"

        assert mock.flush.call_count == 1

    # ----------------------------------------------------------------------
    def test_Stdin(self):
        mock = Mock()

        result = Stream(
            '''python -c "import sys; sys.stdout.write(sys.stdin.read())"''',
            mock,
            stdin="Hello!",
            line_delimited_output=True,
        )

        assert result == 0

        assert mock.write.call_count == 1
        assert mock.write.call_args_list[0].args[0] == "Hello!\n"

        assert mock.flush.call_count == 1

    # ----------------------------------------------------------------------
    def test_AsciiEscape(self):
        mock = Mock()

        result = Stream('''python -c "print('\x1b[31;1mERROR:\x1b[0m Hello!')"''', mock)

        assert result == 0

        assert mock.write.call_count == 16
        assert mock.write.call_args_list[0].args[0] == "\x1b[31;1m"
        assert mock.write.call_args_list[1 + len("ERROR:")].args[0] == "\x1b[0m"
        assert mock.write.call_args_list[15].args[0] == "\n"

        assert mock.flush.call_count == 1

    # ----------------------------------------------------------------------
    def test_Unicode8(self):
        mock = Mock()

        result = Stream('''python -c "import sys; sys.stdout.write('🔥')"''', mock)

        assert result == 0

        assert mock.write.call_count == 1
        assert mock.write.call_args_list[0].args[0] == "🔥"

        assert mock.flush.call_count == 1

    # ----------------------------------------------------------------------
    def test_MultiByte(self):
        mock = Mock()

        result = Stream('''python -c "print('\u2082')"''', mock)

        assert result == 0

        assert mock.write.call_count == 2
        assert mock.write.call_args_list[0].args[0] == "₂"
        assert mock.write.call_args_list[1].args[0] == "\n"

        assert mock.flush.call_count == 1



# ----------------------------------------------------------------------
# |
# |  Capabilities.py
# |
# |  David Brownell <db@DavidBrownell.com>
# |      2023-12-18 15:40:51
# |
# ----------------------------------------------------------------------
# |
# |  Copyright David Brownell 2023-24
# |  Distributed under the MIT License.
# |
# ----------------------------------------------------------------------
"""Contains the Capabilities object"""

import functools
import os
import sys
import textwrap

from typing import Any, Optional, TextIO

from dbrownell_Common.Streams.TextWriter import TextWriter


# ----------------------------------------------------------------------
TextWriterT = TextIO | TextWriter


# ----------------------------------------------------------------------
@functools.total_ordering
class Capabilities:
    """Capabilities of a stream."""

    # ----------------------------------------------------------------------
    # |
    # |  Public Types
    # |
    # ----------------------------------------------------------------------
    DEFAULT_COLUMNS: int = 180

    SIMULATE_TERMINAL_COLUMNS_ENV_VAR: str = "COLUMNS"
    SIMULATE_TERMINAL_INTERACTIVE_ENV_VAR: str = "SIMULATE_TERMINAL_CAPABILITIES_IS_INTERACTIVE"
    SIMULATE_TERMINAL_COLORS_ENV_VAR: str = "SIMULATE_TERMINAL_CAPABILITIES_SUPPORTS_COLORS"
    SIMULATE_TERMINAL_HEADLESS_ENV_VAR: str = "SIMULATE_TERMINAL_CAPABILITIES_IS_HEADLESS"

    # ----------------------------------------------------------------------
    # |
    # |  Public Methods
    # |
    # ----------------------------------------------------------------------
    def __init__(
        self,
        *,
        columns: Optional[int] = None,
        is_headless: Optional[bool] = None,
        is_interactive: Optional[bool] = None,
        supports_colors: Optional[bool] = None,
        stream: Optional[TextWriterT] = None,
        no_column_warning: bool = False,
        ignore_environment: bool = False,
    ):
        # columns
        explicit_columns = False

        if columns is not None:
            explicit_columns = True
        else:
            value = (
                os.getenv(self.__class__.SIMULATE_TERMINAL_COLUMNS_ENV_VAR)
                if not ignore_environment
                else None
            )
            if value is not None:
                columns = int(value)
                explicit_columns = True
            else:
                columns = self.__class__.DEFAULT_COLUMNS
                explicit_columns = True

        # is_interactive
        explicit_is_interactive = False

        if is_interactive is not None:
            explicit_is_interactive = True
        else:
            value = (
                os.getenv(self.__class__.SIMULATE_TERMINAL_INTERACTIVE_ENV_VAR)
                if not ignore_environment
                else None
            )
            if value is not None:
                is_interactive = value != "0"
                explicit_is_interactive = True
            else:
                is_interactive = stream is not None and stream.isatty()

        # is_headless
        explicit_is_headless = False

        if is_headless is not None:
            explicit_is_headless = True
        else:
            value = (
                os.getenv(self.__class__.SIMULATE_TERMINAL_HEADLESS_ENV_VAR)
                if not ignore_environment
                else None
            )
            if value is not None:
                is_headless = value != "0"
                explicit_is_headless = True
            else:
                is_headless = not is_interactive

        # supports_colors
        explicit_supports_colors = False

        if supports_colors is not None:
            explicit_supports_colors = True
        else:
            value = (
                os.getenv(self.__class__.SIMULATE_TERMINAL_COLORS_ENV_VAR) if not ignore_environment else None
            )
            if value is not None:
                supports_colors = value != "0"
                explicit_supports_colors = True
            else:
                supports_colors = is_interactive

        # Commit the values
        self.columns = columns
        self.is_headless = is_headless
        self.is_interactive = is_interactive
        self.supports_colors = supports_colors

        self.explicit_columns = explicit_columns
        self.explicit_is_headless = explicit_is_headless
        self.explicit_is_interactive = explicit_is_interactive
        self.explicit_supports_colors = explicit_supports_colors

        # Validate the settings for sys.stdout
        if stream:
            self.__class__.Set(stream, self)

            if stream is sys.stdout and self.__class__._processed_stdout is False:
                try:
                    import rich
                    from rich.console import ConsoleDimensions

                    if (
                        explicit_columns
                        or explicit_is_headless
                        or explicit_is_interactive
                        or explicit_supports_colors
                    ):
                        rich_args = self._GetRichConsoleArgs()

                        # Width needs to be set separately
                        width_arg = rich_args.pop("width", None)

                        rich.reconfigure(**rich_args)

                        if width_arg is not None:
                            console = rich.get_console()
                            console.size = ConsoleDimensions(width_arg, console.height)

                    # Validate that the width is acceptable. This has to be done AFTER
                    # the capabilities have been associated with the stream.
                    if not no_column_warning:
                        console = rich.get_console()

                        if console.width < self.__class__.DEFAULT_COLUMNS:
                            # Importing here to avoid circular imports
                            from dbrownell_Common import TextwrapEx
                            from dbrownell_Common.Streams.StreamDecorator import (
                                StreamDecorator,
                            )  # pylint: disable=cyclic-import

                            StreamDecorator(
                                sys.stdout,
                                line_prefix=TextwrapEx.CreateWarningPrefix(self.supports_colors),
                            ).write(
                                textwrap.dedent(
                                    """\


                                    Output is configured for a width of '{}', but your terminal has a width of '{}'.

                                    Some formatting may not appear as intended.


                                    """,
                                ).format(
                                    self.__class__.DEFAULT_COLUMNS,
                                    console.width,
                                ),
                            )

                except ImportError:  # pragma: no cover
                    pass  # pragma: no cover

                self.__class__._processed_stdout = True

    # ----------------------------------------------------------------------
    @staticmethod
    def Compare(
        this: "Capabilities",
        that: "Capabilities",
    ) -> int:
        for attribute_data in [
            "columns",
            ("is_headless", True),
            "is_interactive",
            "supports_colors",
        ]:
            if isinstance(attribute_data, tuple):
                attribute_name, invert = attribute_data
            else:
                attribute_name = attribute_data
                invert = False

            this_value = getattr(this, attribute_name)
            that_value = getattr(that, attribute_name)

            if this_value != that_value:
                if invert:
                    return 1 if this_value < that_value else -1

                return -1 if this_value < that_value else 1

        return 0

    # ----------------------------------------------------------------------
    def __lt__(
        self,
        other: "Capabilities",
    ) -> bool:
        return self.Compare(self, other) < 0

    # ----------------------------------------------------------------------
    def __eq__(
        self,
        other: "Capabilities",
    ) -> bool:
        return self.Compare(self, other) == 0

    # ----------------------------------------------------------------------
    @classmethod
    def Get(
        cls,
        stream: TextWriterT,
    ) -> "Capabilities":
        current_capabilities = getattr(stream, cls._EMBEDDED_CAPABILITIES_ATTRIBUTE_NAME, None)
        if current_capabilities is not None:
            return current_capabilities

        return cls(stream=stream)

    # ----------------------------------------------------------------------
    @classmethod
    def Set(
        cls,
        stream: TextWriterT,
        capabilities: "Capabilities",
    ) -> TextWriterT:
        if getattr(stream, cls._EMBEDDED_CAPABILITIES_ATTRIBUTE_NAME, None) is not None:
            raise Exception(
                "Capabilities are assigned to a stream when it is first created and cannot be changed; consider using the `Clone` method."
            )

        setattr(stream, cls._EMBEDDED_CAPABILITIES_ATTRIBUTE_NAME, capabilities)
        return stream

    # ----------------------------------------------------------------------
    def Clone(
        self,
        *,
        columns: Optional[int] = None,
        is_headless: Optional[bool] = None,
        is_interactive: Optional[bool] = None,
        supports_colors: Optional[bool] = None,
    ) -> "Capabilities":
        return self.__class__(
            columns=columns if columns is not None else self.columns,
            is_headless=is_headless if is_headless is not None else self.is_headless,
            is_interactive=is_interactive if is_interactive is not None else self.is_interactive,
            supports_colors=(supports_colors if supports_colors is not None else self.supports_colors),
        )

    # ----------------------------------------------------------------------
    try:
        from rich.console import Console

        def CreateRichConsole(
            self,
            file: Optional[TextWriterT] = None,
            *,
            width: Optional[int] = None,
        ) -> Console:
            """Creates a `rich` `Console` instance."""

            args = self._GetRichConsoleArgs()

            # Width needs to be set separately
            width_arg = width if width is not None else args.pop("width", None)

            args["file"] = file

            from rich.console import Console, ConsoleDimensions

            result = Console(**args)

            if width_arg is not None:
                result.size = ConsoleDimensions(width_arg, result.height)

            return result

    except ImportError:
        # This means that rich wasn't found, which is OK
        pass

    # ----------------------------------------------------------------------
    # |
    # |  Private Types
    # |
    # ----------------------------------------------------------------------
    _EMBEDDED_CAPABILITIES_ATTRIBUTE_NAME: str = "__stream_capabilities"
    _processed_stdout: bool = False

    # ----------------------------------------------------------------------
    # |
    # |  Private Methods
    # |
    # ----------------------------------------------------------------------
    def _GetRichConsoleArgs(self) -> dict[str, Any]:
        """Returns arguments suitable to instantiate a rich `Console` instance."""

        args: dict[str, Any] = {
            "legacy_windows": False,
        }

        if self.explicit_columns:
            args["width"] = self.columns

        if self.explicit_is_interactive:
            args["force_interactive"] = self.is_interactive

        if self.explicit_supports_colors:
            if self.supports_colors:
                # We can't be too aggressive in the selection of a color system
                # or else the content won't display if we over-reach.
                args["color_system"] = "standard"
            else:
                args["no_color"] = True

        return args



# ----------------------------------------------------------------------
# |
# |  PathEx.py
# |
# |  David Brownell <db@DavidBrownell.com>
# |      2024-01-12 23:12:24
# |
# ----------------------------------------------------------------------
# |
# |  Copyright David Brownell 2024
# |  Distributed under the MIT License.
# |
# ----------------------------------------------------------------------
"""Extensions to Path"""

import os
import tempfile

from pathlib import Path, PurePath
from typing import Optional


# ----------------------------------------------------------------------
def CreateTempFileName(
    suffix: Optional[str] = None,
) -> Path:
    filename_handle, filename = tempfile.mkstemp(suffix=suffix)

    os.close(filename_handle)
    os.remove(filename)

    return Path(filename)


# ----------------------------------------------------------------------
def CreateTempDirectory(
    suffix: Optional[str] = None,
) -> Path:
    directory = CreateTempFileName(suffix=suffix)

    directory.mkdir(parents=True, exist_ok=True)

    return directory


# ----------------------------------------------------------------------
def EnsureExists(
    path: Optional[Path],
) -> Path:
    if path is None:
        raise ValueError("Value is None")

    if not path.exists():
        raise ValueError("'{}' does not exist.".format(path))

    return path


# ----------------------------------------------------------------------
def EnsureFile(
    path: Optional[Path],
) -> Path:
    EnsureExists(path)
    assert path is not None

    if not path.is_file():
        raise ValueError("'{}' is not a file.".format(path))

    return path


# ----------------------------------------------------------------------
def EnsureDir(
    path: Optional[Path],
) -> Path:
    EnsureExists(path)
    assert path is not None

    if not path.is_dir():
        raise ValueError("'{}' is not a directory.".format(path))

    return path


# ----------------------------------------------------------------------
def IsDescendant(
    query: PurePath,
    root: PurePath,
) -> bool:
    """Returns True if `query` is a descendant of `root`."""

    try:
        for query_part, root_part in zip(query.parts[: len(root.parts)], root.parts, strict=True):
            if query_part != root_part:
                return False
    except ValueError:
        return False

    return True


# ----------------------------------------------------------------------
def CreateRelativePath(
    origin_path: PurePath,
    dest_path: PurePath,
) -> PurePath:
    """Returns a path to navigate from the origin to the destination."""

    len_origin_path_parts = len(origin_path.parts)
    len_dest_path_parts = len(dest_path.parts)

    # Find all of the path parts that match
    min_length = min(len_origin_path_parts, len_dest_path_parts)

    matching_index = 0

    while matching_index < min_length:
        if origin_path.parts[matching_index] != dest_path.parts[matching_index]:
            break

        matching_index += 1

    relative_path = PurePath(".")

    if matching_index < len_origin_path_parts:
        relative_path = relative_path.joinpath(
            *(
                [
                    "..",
                ]
                * (len_origin_path_parts - matching_index)
            )
        )

    if matching_index < len_dest_path_parts:
        relative_path = relative_path.joinpath(*dest_path.parts[matching_index:])

    return relative_path


# ----------------------------------------------------------------------
def GetCommonPath(
    *path_args: Path,
) -> Optional[Path]:
    paths = [path.resolve() for path in path_args]

    if len(paths) == 1:
        if paths[0].is_dir():
            return paths[0]

        return paths[0].parent

    part_index = 0

    while True:
        is_match = True

        if part_index > len(paths[0].parts) - 1:
            break

        for path in paths[1:]:
            if part_index > len(path.parts) - 1 or path.parts[part_index] != paths[0].parts[part_index]:
                is_match = False
                break

        if not is_match:
            break

        part_index += 1

    if part_index == 0:
        return None

    return Path(*paths[0].parts[:part_index])


# ----------------------------------------------------------------------
def GetSizeDisplay(
    path_or_num_bytes: Path | int,
):
    if isinstance(path_or_num_bytes, Path):
        num_bytes = float(path_or_num_bytes.stat().st_size)
    elif isinstance(path_or_num_bytes, int):
        num_bytes = float(path_or_num_bytes)
    else:
        assert False, path_or_num_bytes  # pragma: no cover

    if num_bytes < 1024.0:
        return "{} B".format(int(num_bytes))

    num_bytes /= 1024.0

    for unit in [
        "K",
        "M",
        "G",
        "T",
        "P",
        "E",
        "Z",
    ]:
        if num_bytes < 1024.0:
            return "{:.1f} {}B".format(num_bytes, unit)

        num_bytes /= 1024.0

    return "{:.1f} YiB".format(num_bytes)


# ----------------------------------------------------------------------
def GetUserDirectory() -> Path:
    if os.name == "nt":
        home_drive = os.getenv("HOMEDRIVE")
        home_path = os.getenv("HOMEPATH")

        if home_drive is not None and home_path is not None:
            return Path(home_drive) / home_path

        assert home_path is not None
        return Path(home_path)

    return Path(os.path.expanduser("~"))



# ----------------------------------------------------------------------
# |
# |  StreamTestHelpers.py
# |
# |  David Brownell <db@DavidBrownell.com>
# |      2024-02-14 10:28:55
# |
# ----------------------------------------------------------------------
# |
# |  Copyright David Brownell 2024
# |  Distributed under the MIT License.
# |
# ----------------------------------------------------------------------
"""Test helpers for content found in ../Streams"""

import os
import re
import sys

from io import StringIO
from typing import Generator, Match, Optional

from dbrownell_Common.Streams.Capabilities import Capabilities
from dbrownell_Common.Streams.DoneManager import DoneManager, Flags as DoneManagerFlags


# ----------------------------------------------------------------------
def ScrubDuration(
    content: str,
    *,
    keep_hours: bool = False,
    keep_minutes: bool = False,
    keep_seconds: bool = False,
) -> str:
    if keep_hours or keep_minutes or keep_seconds:
        # ----------------------------------------------------------------------
        def Replace(
            match: Match,
        ) -> str:
            hours = match.group("hours") if keep_hours else "??"
            minutes = match.group("minutes") if keep_minutes else "??"
            seconds = match.group("seconds") if keep_seconds else "??"

            return f"{hours}:{minutes}:{seconds}"

        # ----------------------------------------------------------------------

        replace_func = Replace
    else:
        replace_func = lambda _: "<scrubbed duration>"

    return re.sub(
        r"""(?#
            Hours                           )(?P<hours>\d+)(?#
            sep                             )\:(?#
            Minutes                         )(?P<minutes>\d+)(?#
            sep                             )\:(?#
            Seconds                         )(?P<seconds>\d+(?:\.\d+)?)(?#
            )""",
        replace_func,
        content,
    )


# ----------------------------------------------------------------------
def GenerateDoneManagerAndContent(
    heading: str = "Heading",
    *,
    expected_result: Optional[int] = None,
    verbose: bool = False,
    debug: bool = False,
    keep_duration_hours: bool = False,
    keep_duration_minutes: bool = False,
    keep_duration_seconds: bool = False,
) -> Generator[DoneManager | str, None, None]:
    """\
    Generates a DoneManager followed by the content populated by the DoneManager.

    Example Usage:
        dm_and_sink = iter(GenerateDoneManagerAndContent())

        dm = cast(DoneManager, next(dm_and_sink))

        dm.WriteInfo("Content")

        content = cast(str, next(dm_and_sink))

        assert content == textwrap.dedent(
            '''\
            Heading...
              INFO: Content
            DONE! (0, <scrubbed duration>)
            ''',
        )
    """

    # Do not decorate the output, regardless of what environment variables specify
    sink = Capabilities.Set(
        StringIO(),
        Capabilities(
            is_headless=True,
            is_interactive=False,
            supports_colors=False,
        ),
    )

    with DoneManager.Create(
        sink,
        heading,
        flags=DoneManagerFlags.Create(verbose=verbose, debug=debug),
    ) as dm:
        yield dm

        final_result = dm.result

    content = ScrubDuration(
        sink.getvalue(),  # type: ignore
        keep_hours=keep_duration_hours,
        keep_minutes=keep_duration_minutes,
        keep_seconds=keep_duration_seconds,
    )

    # Remove any trailing whitespace
    content = "\n".join(line.rstrip() for line in content.split("\n"))

    assert expected_result is None or final_result == expected_result, (
        expected_result,
        final_result,
        content,
    )
    yield content


# ----------------------------------------------------------------------
def InitializeStreamCapabilities(
    stream=sys.stdout,
) -> None:
    """\
    Ensure that the provided stream outputs consistently within a variety of different environments.

    Usage within a test:
        @pytest.fixture(InitializeStreamCapabilities(), scope="session", autouse=True)

    """

    # Initialization can only happen once
    if hasattr(
        stream,
        Capabilities._EMBEDDED_CAPABILITIES_ATTRIBUTE_NAME,  # pylint: disable=protected-access
    ):
        return

    # Override any environment variables that might impact the way in which the output is generated
    os.environ[Capabilities.SIMULATE_TERMINAL_COLUMNS_ENV_VAR] = str(Capabilities.DEFAULT_COLUMNS)
    os.environ[Capabilities.SIMULATE_TERMINAL_INTERACTIVE_ENV_VAR] = "0"
    os.environ[Capabilities.SIMULATE_TERMINAL_COLORS_ENV_VAR] = "1"
    os.environ[Capabilities.SIMULATE_TERMINAL_HEADLESS_ENV_VAR] = "1"

    # Associate the capabilities with the stream
    Capabilities(
        columns=Capabilities.DEFAULT_COLUMNS,
        is_interactive=False,
        is_headless=True,
        supports_colors=True,
        stream=stream,
        no_column_warning=True,
    )



# ----------------------------------------------------------------------
# |
# |  TyperEx.py
# |
# |  David Brownell <db@DavidBrownell.com>
# |      2024-01-15 14:41:06
# |
# ----------------------------------------------------------------------
# |
# |  Copyright David Brownell 2024
# |  Distributed under the MIT License.
# |
# ----------------------------------------------------------------------
"""Extensions to the typer module"""

import re

from dataclasses import dataclass
from enum import auto, Enum
from types import NoneType
from typing import Any, Callable, ClassVar, Iterable, Optional, Type, TypeVar, Union

import typer

from typer import main as typer_main

from dbrownell_Common import TextwrapEx


# ----------------------------------------------------------------------
# |
# |  Public Types
# |
# ----------------------------------------------------------------------
@dataclass(frozen=True)
class TypeDefinitionItem:
    """Information used to generate a dynamic command line argument."""

    # ----------------------------------------------------------------------
    python_type: Type
    parameter_info: typer.models.ParameterInfo

    # ----------------------------------------------------------------------
    @classmethod
    def Create(
        cls,
        python_type: Type,
        *,
        is_optional: bool = False,
        **param_info_kwargs,
    ) -> "TypeDefinitionItem":
        if is_optional or _IsOptionalPythonType(python_type):
            typer_param = typer.Option(None, **param_info_kwargs)
        else:
            typer_param = typer.Argument(..., **param_info_kwargs)

        return cls(python_type, typer_param)


# ----------------------------------------------------------------------
TypeDefinitionItemType = Union[  # pylint: disable=invalid-name
    TypeDefinitionItem,
    # Note that the following values are converted into TypeDefinitionItem instances
    Type,  # Python type annotation
    tuple[
        Type,
        typer.models.ParameterInfo,  # The OptionInfo itself
    ],
]


# ----------------------------------------------------------------------
TypeDefinitionsType = dict[str, TypeDefinitionItemType]


# ----------------------------------------------------------------------
@dataclass(frozen=True)
class PythonCodeGenerator:
    """\
    Object able to generate python code based on the type definitions provided to it.

    Example:
        generator = PythonCodeGenerator.Create(
            {
                "arg1": TypeDefinitionItem(int, typer.Option(10)),
                "arg2": int,
                "arg3": (int, {"help": "This is the help"}),
                "arg4": (int, typer.Option(20)),
            },
            "custom_types_name",
        )

        python_code = textwrap.dedent(
            '''\
            @app.command()
            def DynamicFunc(
                {parameters}
            ):
                print({arguments})

            ''',
        ).format(
            parameters=generator.GenerateParameters(single_line=False),
            arguments=generator.GenerateArguments(single_line=True),
        )
    """

    # ----------------------------------------------------------------------
    class ArgumentTypes(Enum):
        """Indicates how the generation should generate the output."""

        CommaDelimited = auto()
        DictArgs = auto()
        KeywordArgs = auto()

    # ----------------------------------------------------------------------
    DEFAULT_OPTION_TYPE_VALUES_VAR_NAME: ClassVar[str] = "default_parameter_values"

    # ----------------------------------------------------------------------
    python_parameters: dict[str, str]
    python_type_values: dict[str, typer.models.ParameterInfo]

    # ----------------------------------------------------------------------
    @classmethod
    def Create(
        cls,
        type_definitions: TypeDefinitionsType,
        option_type_values_var_name: str = DEFAULT_OPTION_TYPE_VALUES_VAR_NAME,
    ) -> "PythonCodeGenerator":
        parameters: dict[str, str] = {}
        types: dict[str, typer.models.ParameterInfo] = {}

        for k, v in type_definitions.items():
            python_type: None | Type | str = None
            parameter_info: Optional[typer.models.ParameterInfo] = None

            if isinstance(v, TypeDefinitionItem):
                python_type = v.python_type
                parameter_info = v.parameter_info
            elif isinstance(v, tuple):
                python_type, parameter_info = v
            else:
                python_type = v
                parameter_info = typer.Argument()

            assert python_type is not None
            assert parameter_info is not None

            python_type_name: Optional[str] = None

            if isinstance(python_type, str):
                if python_type.startswith("typing."):
                    python_type_name = python_type[len("typing.") :]
                else:
                    python_type_name = python_type
            else:
                python_type_name = python_type.__name__

            assert python_type_name is not None

            types[k] = parameter_info

            parameters[k] = '{name}: {python_type_name}={var_name}["{name}"]'.format(
                name=k,
                python_type_name=python_type_name,
                var_name=option_type_values_var_name,
            )

        return PythonCodeGenerator(parameters, types)

    # ----------------------------------------------------------------------
    def GenerateParameters(
        self,
        *,
        single_line: bool = False,
        indentation: int = 4,
        skip_first_line: bool = True,
    ) -> str:
        if single_line:
            return ", ".join(parameter for parameter in self.python_parameters.values())

        return TextwrapEx.Indent(
            "\n".join("{},".format(parameter) for parameter in self.python_parameters.values()),
            indentation,
            skip_first_line=skip_first_line,
        )

    # ----------------------------------------------------------------------
    def GenerateArguments(
        self,
        *,
        argument_type: "PythonCodeGenerator.ArgumentTypes" = ArgumentTypes.CommaDelimited,
        single_line: bool = False,
        indentation: int = 4,
        skip_first_line: bool = True,
    ) -> str:
        if argument_type == PythonCodeGenerator.ArgumentTypes.CommaDelimited:
            decorate_parameter_func = lambda value: value
        elif argument_type == PythonCodeGenerator.ArgumentTypes.DictArgs:
            decorate_parameter_func = lambda value: '"{value}": {value}'.format(value=value)
        elif argument_type == PythonCodeGenerator.ArgumentTypes.KeywordArgs:
            decorate_parameter_func = lambda value: "{value}={value}".format(value=value)
        else:
            assert False, argument_type  # pragma: no cover

        if single_line:
            return ", ".join(
                decorate_parameter_func(parameter) for parameter in self.python_parameters.keys()
            )

        return TextwrapEx.Indent(
            "\n".join(
                "{},".format(decorate_parameter_func(parameter))
                for parameter in self.python_parameters.keys()
            ),
            indentation,
            skip_first_line=skip_first_line,
        )


# ----------------------------------------------------------------------
# |
# |  Public Functions
# |
# ----------------------------------------------------------------------
def TyperDictArgument(
    type_definitions: TypeDefinitionsType,
    *,
    allow_any__: bool = False,  # Do not produce errors when key values are provided that are not defined in type_definitions
    **argument_info_kwargs,
) -> typer.models.ArgumentInfo:
    """\
    Creates a typer.models.ArgumentInfo instance that is able to process key-value-pairs provided
    on the command line as strings.

    Example:
        CODE:
            @app.command()
            def MyFunc(
                key_value_args: list[str]=TyperDictArgument(
                    ...,
                    dict(
                        one=str,
                        type=int,
                        three=list[int],
                        four=Optional[float],
                        five=(Optional[Path], dict(exists=True, resolve_path=True)),
                    ),
                ),
            ):
                the_dict = PostprocessDictArgument(key_value_args)

                print(the_dict)

        COMMAND LINE:
            MyFunc one:1 two:2 three:3 three:33
                -> {
                    "one": "1",
                    "two": 2,
                    "three": [3, 33],
                    "four": None,
                    "five": None,
                }

            MyFunc one:1 three:3 three: 33
                -> raises error, two is required

            MyFunc one:1 two:2 three:3 three:33 four:1.2345 five:.
                -> {
                    "one": "1",
                    "two": 2,
                    "three": [3, 33],
                    "four": 1.2345,
                    "five": Path("<your working directory here>"),
                }

            MyFunc
                -> raises error, one is required
    """

    return _TyperDictImpl(
        typer.models.ArgumentInfo,
        type_definitions,
        allow_any__,
        None,
        argument_info_kwargs,
    )


# ----------------------------------------------------------------------
def TyperDictOption(
    type_definitions: TypeDefinitionsType,
    *option_info_args,
    allow_any__: bool = False,  # Do not produce errors when key values are provided that are not defined in type_definitions
    **option_info_kwargs,
) -> typer.models.OptionInfo:
    """\
    Creates a typer.models.OptionInfo instance that is able to process key-value-pairs provided
    on the command line as strings.

    Example:
        CODE:
            @app.command()
            def MyFunc(
                key_value_args: list[str]=TyperDictOption(
                    ...,
                    dict(
                        one=str,
                        type=int,
                        three=list[int],
                        four=Optional[float],
                        five=(Optional[Path], dict(exists=True, resolve_path=True)),
                    ),
                ),
            ):
                the_dict = PostprocessDictArgument(key_value_args)

                print(the_dict)

        COMMAND LINE:
            MyFunc one:1 two:2 three:3 three:33
                -> {
                    "one": "1",
                    "two": 2,
                    "three": [3, 33],
                    "four": None,
                    "five": None,
                }

            MyFunc one:1 two:2 three:3 three:33 four:1.2345 five:.
                -> {
                    "one": "1",
                    "two": 2,
                    "three": [3, 33],
                    "four": 1.2345,
                    "five": Path("<your working directory here>"),
                }
    """

    return _TyperDictImpl(
        typer.models.OptionInfo,
        type_definitions,
        allow_any__,
        option_info_args,
        option_info_kwargs,
        force_optional=True,
    )


# ----------------------------------------------------------------------
def PostprocessDictArgument(
    args: Optional[list[Any]],
) -> dict[str, Any]:
    """\
    Converts data that is plumbed through `TyperDictArgument` or `TyperDictOption` into
    a dictionary, as expected.

    This postprocessing step is necessary to ensure that the data (that doesn't look like
    what typer expects it to) makes it through the typer engine mechanics without losing
    anything along the way.

    This function is necessary because the entire solution is a bit of a hack.
    """

    if args is None:
        return {}

    assert len(args) == 1
    assert isinstance(args[0], dict)
    return args[0]


# ----------------------------------------------------------------------
def ProcessDynamicArgs(
    ctx: typer.Context,
    type_definitions: TypeDefinitionsType,
) -> dict[str, Any]:
    """\
    Process arguments that have been dynamically generated.

    Example:
        app = typer.Typer()

        @app.command(
            name="MyFunc",
            context_settings={"allow_extra_args": True, "ignore_unknown_options": True},
        )
        def MyFunc(
            ctx: typer.Context,
            arg1: int=typer.Argument(...),
            arg2: bool=typer.Option(False, "--arg2"),
        ) -> None:
            # Normally, the following arguments would be generated dynamically
            type_definitions = {
                "extra_args1": (int, dict(min=10, max=100)),
                "extra_args2": str,
                "extra_args3": (list[int], typer.Option(None, "--extra-arg3")),
            }

            extra_args = ProcessDynamicArgs(ctx, type_definitions)
    """

    # Group the arguments
    arguments: list[tuple[str, Optional[str]]] = []

    for arg in ctx.args:
        if arg.startswith("-"):
            arg = arg.lstrip("-")

            arguments.append((arg, None))
        else:
            # The value should be associated with a keyword encountered prior
            if not arguments or arguments[-1][-1] is not None:
                raise typer.BadParameter("Got unexpected extra argument ({})".format(arg))

            arguments[-1] = (arguments[-1][0], arg)

    # Read default_map, which may be populated by typer-config
    for k, v in (ctx.default_map or {}).items():
        arguments.append((k, v))

    # Invoke the dynamic functionality
    return _ProcessArgumentsImpl(
        _TypeDefinitionItemsToClickParams(
            ResolveTypeDefinitions(
                type_definitions,
                force_optional=True,
            ),
        ),
        arguments,
        ctx=ctx,
    )


# ----------------------------------------------------------------------
def ResolveTypeDefinitions(
    type_definitions: TypeDefinitionsType,
    *,
    force_optional: bool = False,
) -> dict[str, TypeDefinitionItem]:
    # ----------------------------------------------------------------------
    def ResolveItem(
        item: TypeDefinitionItemType,
    ) -> TypeDefinitionItem:
        type_definition_item: Optional[TypeDefinitionItem] = None

        if isinstance(item, TypeDefinitionItem):
            type_definition_item = item
        elif isinstance(item, tuple):
            python_type, parameter_info = item

            type_definition_item = TypeDefinitionItem(python_type, parameter_info)
        else:
            type_definition_item = TypeDefinitionItem.Create(
                item,
                is_optional=force_optional,
            )

        assert type_definition_item is not None

        if force_optional and not isinstance(type_definition_item.parameter_info, typer.models.OptionInfo):
            raise Exception("Optional types must be defined as typer.Option instances.")

        return type_definition_item

    # ----------------------------------------------------------------------

    results: dict[str, TypeDefinitionItem] = {}

    for k, v in type_definitions.items():
        results[k] = ResolveItem(v)

    return results


# ----------------------------------------------------------------------
# |
# |  Private Functions
# |
# ----------------------------------------------------------------------
def _TypeDefinitionItemsToClickParams(
    type_definitions: dict[str, TypeDefinitionItem],
) -> dict[str, tuple[Any, Callable[..., Any]]]:
    return {
        k: typer_main.get_click_param(
            typer.models.ParamMeta(
                name=k,
                default=v.parameter_info,
                annotation=v.python_type,
            ),
        )
        for k, v in type_definitions.items()
    }


# ----------------------------------------------------------------------
_TyperT = TypeVar("_TyperT", typer.models.ArgumentInfo, typer.models.OptionInfo)
_TyperDictImplRegex = re.compile(
    r"""(?#
    Start of Line                           )^(?#
    Key                                     )(?P<key>(?:\\[:=]|[^:=])+)(?#
    Optional Value Begin                    )(?:(?#
        Sep                                 )\s*[:=]\s*(?#
        Value                               )(?P<value>.+)(?#
    Optional Value End                      ))?(?#
    End of Line                             )$(?#
    )""",
)


def _TyperDictImpl(
    typer_type: Type[_TyperT],
    type_definitions: TypeDefinitionsType,
    allow_unknown: bool,
    args: Optional[tuple[str, ...]],
    kwargs: dict[str, Any],
    *,
    force_optional: bool = False,
) -> _TyperT:
    click_params = _TypeDefinitionItemsToClickParams(
        ResolveTypeDefinitions(
            type_definitions,
            force_optional=force_optional,
        ),
    )

    # Prepare the result
    original_callback = kwargs.pop("callback", None)

    # ----------------------------------------------------------------------
    def NewCallback(
        ctx: typer.Context,
        param: typer.CallbackParam,
        values: list[str],
    ) -> list[dict[str, Any]]:
        arguments: list[tuple[str, Optional[str]]] = []

        for value in values:
            match = _TyperDictImplRegex.match(value)
            if not match:
                raise typer.BadParameter(
                    "'{}' is not a valid dictionary parameters; expected '<key>:<value>' or '<key> = <value>'.".format(
                        value,
                    ),
                )

            key = match.group("key").replace("\\:", ":").replace("\\=", "=")
            value = (match.group("value") or "").replace("\\:", ":").replace("\\=", "=")

            arguments.append((key, value))

        results = _ProcessArgumentsImpl(
            click_params,
            arguments,
            ctx=ctx,
            allow_unknown=allow_unknown,
        )

        # Invoke the original callback
        if original_callback:
            original_callback(ctx, param, results)

        # Convert the results to a list to make it through the typer plumbing
        # unchanged (since the original type was decorated as a list)
        return [
            results,
        ]

    # ----------------------------------------------------------------------

    return typer_type(
        **{
            **kwargs,
            **{
                "default": ...,
                "callback": NewCallback,
                "param_decls": args,
            },
        },
    )


# ----------------------------------------------------------------------
def _ProcessArgumentsImpl(
    click_params: dict[str, tuple[Any, Callable[..., Any]]],
    arguments: Iterable[tuple[str, Optional[str]]],
    *,
    ctx: Optional[typer.Context],
    allow_unknown: bool = False,
) -> dict[str, Any]:
    # Create information to map from the argument keyword to the result name
    argument_to_result_names: dict[str, str] = {}

    for result_name, (click_param, _) in click_params.items():
        for opt in click_param.opts:
            if opt.startswith("--"):
                opt = opt[2:]

            assert opt not in argument_to_result_names, opt
            argument_to_result_names[opt] = result_name

    results: dict[str, Any] = {}

    # Group the argument values
    for key, value in arguments:
        result_name = argument_to_result_names.get(key, None)  # type: ignore
        if result_name is None:
            if not allow_unknown:
                if not click_params:
                    error_context = "custom arguments are not supported"
                else:
                    error_context = "valid keys are {}".format(
                        ", ".join("'{}'".format(name) for name in argument_to_result_names),
                    )

                raise typer.BadParameter("'{}' is not a valid key; {}.".format(key, error_context))

            result_name = key

        result_value = results.get(result_name, None)

        if result_value is not None:
            if isinstance(result_value, list):
                result_value.append(value)
            else:
                results[result_name] = [result_value, value]
        else:
            results[result_name] = value

    # Convert the values
    does_not_exist = object()

    for param_name, param_info in click_params.items():
        # For some reason, typer uses 2 different techniques to indicate that a type
        # is a list (one technique is used with Arguments, the other is used for Options).
        is_list = param_info[0].nargs == -1 or param_info[0].multiple

        param_results = results.get(param_name, does_not_exist)
        if param_results is does_not_exist:
            if param_info[0].required:
                param_results = None
            else:
                if param_info[0].default is None:
                    param_results = [] if is_list else None
                else:
                    results[param_name] = param_info[0].default

                continue

        if param_results is None:
            if isinstance(param_info[0].default, bool):
                param_results = not param_info[0].default
            else:
                raise typer.BadParameter("A value must be provided for '{}'.".format(param_name))

        if is_list:
            if not isinstance(param_results, list):
                param_results = [
                    param_results,
                ]
        else:
            if isinstance(param_results, list):
                # Take the last value
                param_results = param_results[-1]

        param_results = param_info[0].process_value(ctx, param_results)

        if param_info[1] is not None:
            param_results = param_info[1](param_results)

        results[param_name] = param_results

    return results


# ----------------------------------------------------------------------
def _IsOptionalPythonType(
    python_type: Type,
) -> bool:
    return (
        # The type was defined as a union and one of the union values is None
        getattr(python_type, "__origin__", None) is Union
        and any(value is NoneType for value in python_type.__args__)  # type: ignore
    )



# ----------------------------------------------------------------------
# |
# |  PathEx_UnitTest.py
# |
# |  David Brownell <db@DavidBrownell.com>
# |      2024-01-12 23:13:51
# |
# ----------------------------------------------------------------------
# |
# |  Copyright David Brownell 2024
# |  Distributed under the MIT License.
# |
# ----------------------------------------------------------------------
"""Unit tests for PathEx.py"""

import re

from pathlib import Path

import pytest

from dbrownell_Common.PathEx import *


# ----------------------------------------------------------------------
def test_CreateTempFilename():
    temp_filename = CreateTempFileName()

    assert temp_filename.suffix == ""
    assert not temp_filename.exists()
    assert temp_filename.parent.is_dir()

    temp_filename = CreateTempFileName(".txt")

    assert temp_filename.suffix == ".txt"


# ----------------------------------------------------------------------
def test_CreateTempDirectory():
    temp_directory = CreateTempDirectory()

    assert temp_directory.is_dir()


# ----------------------------------------------------------------------
def test_EnsureExists():
    this_file = Path(__file__)

    # Exists
    assert EnsureExists(this_file) == this_file

    # None
    with pytest.raises(
        ValueError,
        match=re.escape("Value is None"),
    ):
        EnsureExists(None)

    # Does not exist
    filename = this_file.parent / "Does Not Exist.txt"

    with pytest.raises(
        ValueError,
        match=re.escape("'{}' does not exist.".format(filename)),
    ):
        EnsureExists(filename)


# ----------------------------------------------------------------------
def test_EnsureFile():
    this_file = Path(__file__)

    # Exists
    assert EnsureFile(this_file) == this_file

    # None
    with pytest.raises(
        ValueError,
        match=re.escape("Value is None"),
    ):
        EnsureFile(None)

    # Does not exist
    filename = this_file.parent / "Does Not Exist.txt"

    with pytest.raises(
        ValueError,
        match=re.escape("'{}' does not exist.".format(filename)),
    ):
        EnsureFile(filename)

    # Not a file
    with pytest.raises(
        ValueError,
        match=re.escape("'{}' is not a file.".format(this_file.parent)),
    ):
        EnsureFile(this_file.parent)


# ----------------------------------------------------------------------
def test_EnsureDir():
    this_file = Path(__file__)

    # Exists
    assert EnsureDir(this_file.parent) == this_file.parent

    # None
    with pytest.raises(
        ValueError,
        match=re.escape("Value is None"),
    ):
        EnsureDir(None)

    # Does not exist
    filename = this_file.parent / "Does Not Exist"

    with pytest.raises(
        ValueError,
        match=re.escape("'{}' does not exist.".format(filename)),
    ):
        EnsureDir(filename)

    # Not a directory
    with pytest.raises(
        ValueError,
        match=re.escape("'{}' is not a directory.".format(this_file)),
    ):
        EnsureDir(this_file)


# ----------------------------------------------------------------------
def test_IsDescendant():
    assert IsDescendant(Path("/a/b/c"), Path("/a/b/c")) is True
    assert IsDescendant(Path("/a/b/c"), Path("/x/y/z")) is False
    assert IsDescendant(Path("/a/b/c/d/e/f"), Path("/a/b/c")) is True
    assert IsDescendant(Path("/a/b"), Path("/a/b/c")) is False


# ----------------------------------------------------------------------
def test_CreateRelativePath():
    assert CreateRelativePath(PurePath("/a/b"), PurePath("/a/c")) == PurePath("../c")
    assert CreateRelativePath(PurePath("/a/b/c"), PurePath("/a/b/d/e")) == PurePath("../d/e")
    assert CreateRelativePath(PurePath("/a/b/c"), PurePath("/x/y/z")) == PurePath("../../../x/y/z")


# ----------------------------------------------------------------------
def test_GetCommonPath():
    parent_dir = Path(__file__).parent

    assert GetCommonPath(parent_dir) == parent_dir
    assert GetCommonPath(Path(__file__)) == parent_dir
    assert GetCommonPath(Path("a/b/c"), Path("a/b/d")) == Path("a/b").resolve()
    assert GetCommonPath(Path("a/b/c"), Path("a/b")) == Path("a/b").resolve()
    assert (
        GetCommonPath(Path("a/b/c"), Path("1/2/3")) == Path.cwd()
    )  # This happens because the paths are resolved before processing
    assert GetCommonPath(Path("a/b/c"), Path("a/b/c/d")) == Path("a/b/c").resolve()

    if os.name == "nt":
        assert GetCommonPath(Path(r"C:\a\b\c"), Path(r"D:\1\2\3")) is None
    else:
        assert GetCommonPath(Path("/a/b/c"), Path("/1/2/3")) == Path("/")


# ----------------------------------------------------------------------
def test_GetSizeDisplay():
    assert GetSizeDisplay(1000) == "1000 B"
    assert GetSizeDisplay(10000) == "9.8 KB"
    assert GetSizeDisplay(100000) == "97.7 KB"
    assert GetSizeDisplay(1000000) == "976.6 KB"
    assert GetSizeDisplay(10000000) == "9.5 MB"
    assert GetSizeDisplay(100000000) == "95.4 MB"
    assert GetSizeDisplay(1000000000) == "953.7 MB"
    assert GetSizeDisplay(10000000000) == "9.3 GB"
    assert GetSizeDisplay(100000000000000000) == "88.8 PB"
    assert GetSizeDisplay(1000000000000000000000) == "867.4 EB"
    assert GetSizeDisplay(10000000000000000000000000) == "8.3 YiB"
    assert GetSizeDisplay(1000000000000000000000000000000) == "827180.6 YiB"
    assert GetSizeDisplay(Path(__file__)) != ""


# ----------------------------------------------------------------------
def test_GetUserDirectory():
    result = GetUserDirectory()

    assert result.is_dir(), result



# ----------------------------------------------------------------------
# |
# |  StreamDecorator_UnitTest.py
# |
# |  David Brownell <db@DavidBrownell.com>
# |      2024-01-05 15:03:43
# |
# ----------------------------------------------------------------------
# |
# |  Copyright David Brownell 2024
# |  Distributed under the MIT License.
# |
# ----------------------------------------------------------------------
"""Unit tests for StreamDecorator.py."""

import re

from typing import Optional

import pytest

from dbrownell_Common.Streams.StreamDecorator import StreamDecorator
from dbrownell_Common.Streams.TextWriter import TextWriter


# ----------------------------------------------------------------------
# |
# |  Public Types
# |
# ----------------------------------------------------------------------
class FakeStream(TextWriter):
    # ----------------------------------------------------------------------
    def __init__(
        self,
        fileno: int,
        isatty: bool = False,
    ):
        self.content: list[Optional[str]] = []
        self._fileno = fileno
        self._isatty = isatty

    # ----------------------------------------------------------------------
    def isatty(self) -> bool:
        return self._isatty

    # ----------------------------------------------------------------------
    def write(
        self,
        content: str,
    ) -> int:
        self.content.append(content)
        return len(content)

    # ----------------------------------------------------------------------
    def flush(self) -> None:
        self.content.append(None)

    # ----------------------------------------------------------------------
    def fileno(self) -> int:
        return self._fileno

    # ----------------------------------------------------------------------
    def close(self) -> None:
        pass


# ----------------------------------------------------------------------
# |
# |  Public Functions
# |
# ----------------------------------------------------------------------
def test_Empty():
    stream = StreamDecorator(None)

    assert stream.col_offset == 0
    assert stream.has_stdout is False
    assert stream.has_streams is False
    assert stream.wrote_content is False
    assert stream.isatty() is False
    assert stream.fileno() == 0

    stream.write("test")
    assert stream.col_offset == len("test")
    assert stream.wrote_content is True

    stream.flush()
    assert stream.col_offset == len("test")

    assert list(stream.EnumStreams()) == []

    assert stream.GetLinePrefix(0) == ""
    assert stream.GetCompleteLinePrefix() == ""

    stream.close()


# ----------------------------------------------------------------------
def test_SingleSimpleStream():
    fake = FakeStream(123)
    stream = StreamDecorator(fake)

    assert stream.col_offset == 0
    assert stream.has_stdout is False
    assert stream.has_streams is True
    assert stream.wrote_content is False
    assert stream.isatty() is False
    assert stream.fileno() == fake.fileno()
    assert fake.content == []

    assert list(stream.EnumStreams()) == [
        fake,
    ]
    assert stream.GetLinePrefix(0) == ""
    assert stream.GetCompleteLinePrefix() == ""

    # "test"
    stream.write("test")
    assert stream.col_offset == len("test")
    assert stream.wrote_content is True
    assert fake.content == [
        "test",
    ]

    # "\n"
    stream.write("\n")
    assert stream.col_offset == 0
    assert stream.wrote_content is True
    assert fake.content == ["test", "\n"]

    # "test2"
    stream.write("test2")
    assert stream.col_offset == len("test2")
    assert stream.wrote_content is True
    assert fake.content == [
        "test",
        "\n",
        "test2",
    ]

    # flush
    stream.flush()
    assert stream.col_offset == len("test2")
    assert stream.wrote_content is True
    assert fake.content == ["test", "\n", "test2", None]

    # ""
    stream.write("")
    assert stream.col_offset == len("test2")
    assert fake.content == ["test", "\n", "test2", None]

    # "abc\nd"
    stream.write("abc\nd")
    assert stream.col_offset == 1
    assert fake.content == ["test", "\n", "test2", None, "abc", "\n", "d"]

    # "e\nfgh\n"
    stream.write("e\nfgh\n")
    assert stream.col_offset == 0
    assert fake.content == ["test", "\n", "test2", None, "abc", "\n", "d", "e", "\n", "fgh", "\n"]


# ----------------------------------------------------------------------
def test_SingleStreamWithLinePrefixAndSuffix():
    fake = FakeStream(123)
    stream = StreamDecorator(
        fake,
        "__prefix__",
        "__suffix__",
    )

    stream.write("Hello world!\n")
    assert stream.col_offset == 0
    assert stream.wrote_content is True
    assert fake.content == ["__prefix__", "Hello world!", "__suffix__", "\n"]

    stream.write("more")
    assert stream.col_offset == len("__prefix__more")
    assert stream.wrote_content is True
    assert fake.content == ["__prefix__", "Hello world!", "__suffix__", "\n", "__prefix__", "more"]

    stream.flush()
    assert stream.col_offset == len("__prefix__more")
    assert stream.wrote_content is True
    assert fake.content == [
        "__prefix__",
        "Hello world!",
        "__suffix__",
        "\n",
        "__prefix__",
        "more",
        None,
    ]

    stream.close()
    assert fake.content == [
        "__prefix__",
        "Hello world!",
        "__suffix__",
        "\n",
        "__prefix__",
        "more",
        None,
        None,
    ]


# ----------------------------------------------------------------------
def test_SingleStreamWithSuffixFunctor():
    fake = FakeStream(123)
    stream = StreamDecorator(
        fake,
        line_suffix=lambda x: "__suffix__({})".format(x),
    )

    stream.write("one\n")
    assert stream.col_offset == 0
    assert fake.content == ["one", "__suffix__(3)", "\n"]

    stream.write("more\n")
    assert stream.col_offset == 0
    assert fake.content == ["one", "__suffix__(3)", "\n", "more", "__suffix__(4)", "\n"]


# ----------------------------------------------------------------------
def test_SingleComplicatedStream():
    fake = FakeStream(123)

    stream = StreamDecorator(
        fake,
        "__prefix__",
        "__suffix__",
        "STREAM\nHEADER\n",
        "STREAM\nFOOTER\n",
    )

    assert stream.col_offset == 0
    assert fake.content == []

    stream.write("Hello world!")
    assert stream.col_offset == len("__prefix__Hello world!")
    assert fake.content == ["STREAM\nHEADER\n", "__prefix__", "Hello world!"]

    stream.write("\nWith Newline\n")
    assert stream.col_offset == 0
    assert fake.content == [
        "STREAM\nHEADER\n",
        "__prefix__",
        "Hello world!",
        "__suffix__",
        "\n",
        "__prefix__",
        "With Newline",
        "__suffix__",
        "\n",
    ]

    stream.write("no newline before suffix")
    assert stream.col_offset == len("__prefix__no newline before suffix")
    assert fake.content == [
        "STREAM\nHEADER\n",
        "__prefix__",
        "Hello world!",
        "__suffix__",
        "\n",
        "__prefix__",
        "With Newline",
        "__suffix__",
        "\n",
        "__prefix__",
        "no newline before suffix",
    ]

    stream.close()
    assert stream.col_offset == 0
    assert fake.content == [
        "STREAM\nHEADER\n",
        "__prefix__",
        "Hello world!",
        "__suffix__",
        "\n",
        "__prefix__",
        "With Newline",
        "__suffix__",
        "\n",
        "__prefix__",
        "no newline before suffix",
        "STREAM\nFOOTER\n",
        None,
    ]


# ----------------------------------------------------------------------
def test_ExceptionsOnClosed():
    fake = FakeStream(123)

    stream = StreamDecorator(fake)
    stream.close()

    exception = re.escape("Instance is closed.")

    with pytest.raises(Exception, match=exception):
        stream.write("test")

    with pytest.raises(Exception, match=exception):
        stream.flush()

    with pytest.raises(Exception, match=exception):
        stream.close()


# ----------------------------------------------------------------------
def test_MultipleStreams():
    fake1 = FakeStream(123)
    fake2 = FakeStream(456)

    stream = StreamDecorator([fake1, fake2], line_prefix="# ")

    stream.write("Hello world!\n")
    assert fake1.content == ["# ", "Hello world!", "\n"]
    assert fake2.content == ["# ", "Hello world!", "\n"]


# ----------------------------------------------------------------------
def test_GetLinePrefix():
    assert StreamDecorator(FakeStream(123)).GetLinePrefix(0) == ""
    assert (
        StreamDecorator(FakeStream(123), line_prefix=lambda x: "__prefix__({})".format(x)).GetLinePrefix(0)
        == "__prefix__(0)"
    )


# ----------------------------------------------------------------------
def test_DecorateEmptyLines():
    fake = FakeStream(123)
    stream = StreamDecorator(
        fake,
        "__prefix__",
        "__suffix__",
        decorate_empty_lines=False,
    )

    stream.write("Hello\n\nWorld\n")
    assert stream.col_offset == 0
    assert fake.content == [
        "__prefix__",
        "Hello",
        "__suffix__",
        "\n",
        "\n",
        "__prefix__",
        "World",
        "__suffix__",
        "\n",
    ]

    fake = FakeStream(123)

    stream = StreamDecorator(
        fake,
        "__prefix__",
        "__suffix__",
        decorate_empty_lines=True,
    )

    stream.write("Hello\n\nWorld\n")
    assert stream.col_offset == 0
    assert fake.content == [
        "__prefix__",
        "Hello",
        "__suffix__",
        "\n",
        "__prefix__",
        "__suffix__",
        "\n",
        "__prefix__",
        "World",
        "__suffix__",
        "\n",
    ]


# ----------------------------------------------------------------------
def test_GetCompleteLinePrefixNoNested():
    stream = StreamDecorator(FakeStream(123), "__prefix__")

    assert stream.GetCompleteLinePrefix() == "__prefix__"
    assert stream.GetCompleteLinePrefix(include_self=False) == ""


# ----------------------------------------------------------------------
def test_Nested():
    fake = FakeStream(123)
    stream = StreamDecorator(
        StreamDecorator(
            fake,
            "__prefix__(inner)",
            "__suffix__(inner)",
        ),
        "__prefix__(outer)",
        "__suffix__(outer)",
    )

    assert stream.GetCompleteLinePrefix() == "__prefix__(inner)__prefix__(outer)"
    assert stream.GetCompleteLinePrefix(include_self=False) == "__prefix__(inner)"

    assert stream.col_offset == 0
    assert fake.content == []

    stream.write("Hello world!\nabc")
    assert stream.col_offset == len("__prefix__(outer)abc")
    assert fake.content == [
        "__prefix__(inner)",
        "__prefix__(outer)",
        "Hello world!",
        "__suffix__(outer)",
        "__suffix__(inner)",
        "\n",
        "__prefix__(inner)",
        "__prefix__(outer)",
        "abc",
    ]



# ----------------------------------------------------------------------
# |
# |  StreamTestHelpers_UnitTest.py
# |
# |  David Brownell <db@DavidBrownell.com>
# |      2024-02-14 10:40:21
# |
# ----------------------------------------------------------------------
# |
# |  Copyright David Brownell 2024
# |  Distributed under the MIT License.
# |
# ----------------------------------------------------------------------
"""Unit tests for StreamTestHelpers.py"""

import textwrap

from typing import cast

import pytest

from dbrownell_Common.TestHelpers.StreamTestHelpers import *


# ----------------------------------------------------------------------
class TestScrubDuration:
    # ----------------------------------------------------------------------
    def test_Standard(self):
        assert ScrubDuration(
            textwrap.dedent(
                """\
                    1:23:45
                    22:33:44
                    """,
            ),
        ) == textwrap.dedent(
            """\
                <scrubbed duration>
                <scrubbed duration>
                """,
        )

    # ----------------------------------------------------------------------
    def test_KeepHours(self):
        assert ScrubDuration(
            textwrap.dedent(
                """\
                    1:23:45
                    22:33:44
                    """,
            ),
            keep_hours=True,
        ) == textwrap.dedent(
            """\
                1:??:??
                22:??:??
                """,
        )

    # ----------------------------------------------------------------------
    def test_KeepMinutes(self):
        assert ScrubDuration(
            textwrap.dedent(
                """\
                    1:23:45
                    22:33:44
                    """,
            ),
            keep_minutes=True,
        ) == textwrap.dedent(
            """\
                ??:23:??
                ??:33:??
                """,
        )

    # ----------------------------------------------------------------------
    def test_KeepSeconds(self):
        assert ScrubDuration(
            textwrap.dedent(
                """\
                    1:23:45
                    22:33:44
                    """,
            ),
            keep_seconds=True,
        ) == textwrap.dedent(
            """\
                ??:??:45
                ??:??:44
                """,
        )

    # ----------------------------------------------------------------------
    def test_KeepAll(self):
        assert ScrubDuration(
            textwrap.dedent(
                """\
                    1:23:45
                    22:33:44
                    """,
            ),
            keep_hours=True,
            keep_minutes=True,
            keep_seconds=True,
        ) == textwrap.dedent(
            """\
                1:23:45
                22:33:44
                """,
        )


# ----------------------------------------------------------------------
class TestGenerateDoneManagerAndContent:
    # ----------------------------------------------------------------------
    def test_Standard(self):
        dm_and_content = iter(GenerateDoneManagerAndContent())

        dm = cast(DoneManager, next(dm_and_content))

        dm.WriteInfo("Content")

        content = cast(str, next(dm_and_content))

        assert content == textwrap.dedent(
            """\
            Heading...
              INFO: Content
            DONE! (0, <scrubbed duration>)
            """,
        )

    # ----------------------------------------------------------------------
    def test_CustomHeading(self):
        dm_and_content = iter(GenerateDoneManagerAndContent("Custom Heading"))

        dm = cast(DoneManager, next(dm_and_content))

        dm.WriteLine("Line 1")

        content = cast(str, next(dm_and_content))

        assert content == textwrap.dedent(
            """\
            Custom Heading...
              Line 1
            DONE! (0, <scrubbed duration>)
            """,
        )

    # ----------------------------------------------------------------------
    def test_KeepHours(self):
        dm_and_content = iter(GenerateDoneManagerAndContent(keep_duration_hours=True))

        dm = cast(DoneManager, next(dm_and_content))

        dm.WriteInfo("Content")

        content = cast(str, next(dm_and_content))

        assert content == textwrap.dedent(
            """\
            Heading...
              INFO: Content
            DONE! (0, 0:??:??)
            """,
        )

    # ----------------------------------------------------------------------
    def test_KeepMinutes(self):
        dm_and_content = iter(GenerateDoneManagerAndContent(keep_duration_minutes=True))

        dm = cast(DoneManager, next(dm_and_content))

        dm.WriteInfo("Content")

        content = cast(str, next(dm_and_content))

        assert content == textwrap.dedent(
            """\
            Heading...
              INFO: Content
            DONE! (0, ??:00:??)
            """,
        )

    # ----------------------------------------------------------------------
    def test_ExpectedResult(self):
        expected_result = 123

        # ----------------------------------------------------------------------
        def Impl(
            provided_result: int,
        ):
            dm_and_content = iter(GenerateDoneManagerAndContent(expected_result=expected_result))

            dm = cast(DoneManager, next(dm_and_content))

            dm.result = provided_result

            return cast(str, next(dm_and_content))

        # ----------------------------------------------------------------------

        # Test 1
        with pytest.raises(AssertionError):
            Impl(0)

        # Test 2
        assert Impl(expected_result) == textwrap.dedent(
            """\
            Heading...DONE! (123, <scrubbed duration>)
            """,
        )

    # ----------------------------------------------------------------------
    def test_WriteVerbose(self):
        # ----------------------------------------------------------------------
        def Impl(
            verbose: bool,
        ) -> str:
            dm_and_content = iter(GenerateDoneManagerAndContent(verbose=verbose))

            dm = cast(DoneManager, next(dm_and_content))

            dm.WriteVerbose("Line 1")

            return cast(str, next(dm_and_content))

        # ----------------------------------------------------------------------

        assert Impl(False) == textwrap.dedent(
            """\
            Heading...DONE! (0, <scrubbed duration>)
            """,
        )

        assert Impl(True) == textwrap.dedent(
            """\
            Heading...
              VERBOSE: Line 1
            DONE! (0, <scrubbed duration>)
            """,
        )


# ----------------------------------------------------------------------
def test_InitializeStreamCapabilites():
    # It should be safe to call this multiple times
    InitializeStreamCapabilities()
    InitializeStreamCapabilities()

    assert True



# ----------------------------------------------------------------------
# |
# |  Capabilities_UnitTest.py
# |
# |  David Brownell <db@DavidBrownell.com>
# |      2023-12-31 21:35:58
# |
# ----------------------------------------------------------------------
# |
# |  Copyright David Brownell 2023-24
# |  Distributed under the MIT License.
# |
# ----------------------------------------------------------------------
"""Unit tests for Capabilities.py."""

import re
import sys

from unittest.mock import MagicMock as Mock

import pytest

from dbrownell_Common.Streams.Capabilities import *
from dbrownell_Common.Streams.StreamDecorator import StreamDecorator


# ----------------------------------------------------------------------
def test_DefaultConstruct():
    c = Capabilities(ignore_environment=True)

    assert c.columns == Capabilities.DEFAULT_COLUMNS
    assert c.is_headless is True
    assert c.is_interactive is False
    assert c.supports_colors is False

    assert c.explicit_columns is True
    assert c.explicit_is_headless is False
    assert c.explicit_is_interactive is False
    assert c.explicit_supports_colors is False


# ----------------------------------------------------------------------
def test_Initialize():
    columns = Mock()
    is_headless = Mock()
    is_interactive = Mock()
    supports_colors = Mock()

    c = Capabilities(
        columns=columns,
        is_headless=is_headless,
        is_interactive=is_interactive,
        supports_colors=supports_colors,
    )

    assert c.columns is columns
    assert c.is_headless is is_headless
    assert c.is_interactive is is_interactive
    assert c.supports_colors is supports_colors
    assert c.explicit_columns is True
    assert c.explicit_is_headless is True
    assert c.explicit_is_interactive is True
    assert c.explicit_supports_colors is True


# ----------------------------------------------------------------------
def test_WithStdout():
    # I'm not sure how to temporarily redirect sys.stdout here without screwing
    # up other tests running concurrently.
    assert Capabilities._processed_stdout is False

    columns = 100

    # Note that this will generate a warning
    c = Capabilities(
        stream=sys.stdout,
        columns=columns,
    )

    assert Capabilities._processed_stdout is True
    assert c.columns == columns


# ----------------------------------------------------------------------
def test_SetError():
    # Capabilities are associated with a StreamDecorator when it is created.
    stream = StreamDecorator(None)

    with pytest.raises(
        Exception,
        match=re.escape(
            "Capabilities are assigned to a stream when it is first created and cannot be changed; consider using the `Clone` method."
        ),
    ):
        Capabilities(stream=stream)


# ----------------------------------------------------------------------
def test_Clone():
    original_capabilities = Capabilities(is_headless=True)
    new_capabilities = original_capabilities.Clone()

    assert new_capabilities is not original_capabilities
    assert new_capabilities == original_capabilities


# ----------------------------------------------------------------------
def test_Compare():
    assert Capabilities() == Capabilities()
    assert Capabilities(ignore_environment=True) != Capabilities(is_headless=False)
    assert Capabilities(ignore_environment=True) < Capabilities(is_headless=False)
    assert Capabilities(is_headless=False) > Capabilities(ignore_environment=True)
    assert Capabilities(is_headless=False, is_interactive=True, supports_colors=False) < Capabilities(
        is_headless=False, is_interactive=True, supports_colors=True
    )



# ----------------------------------------------------------------------
# |
# |  Types.py
# |
# |  David Brownell <db@DavidBrownell.com>
# |      2024-01-19 19:59:41
# |
# ----------------------------------------------------------------------
# |
# |  Copyright David Brownell 2024
# |  Distributed under the MIT License.
# |
# ----------------------------------------------------------------------
"""Common type-related functionality"""

import sys

from typing import Optional, TypeVar


# ----------------------------------------------------------------------
# |
# |  Public Types
# |
# ----------------------------------------------------------------------
TypeT = TypeVar("TypeT")


# ----------------------------------------------------------------------
# |
# |  Public Functions
# |
# ----------------------------------------------------------------------
def extension(func):  # pylint: disable=invalid-name
    """\
    Decorator that indicates that the method is a method that is intended to be extended by derived
    classes to override functionality if necessary.

    This decorator does not add any functionality, but serves as documentation that communicates
    intentions behind how the class is intended to be used.
    """

    return func


# ----------------------------------------------------------------------
if sys.version_info[0] > 3 and sys.version_info[1] >= 12:
    from typing import override as override_impl

    override = override_impl
else:

    def override(func):  # pylint: disable=invalid-name
        """\
        Decorator that indicates that the method is a method that overrides an abstract- or extension-
        method in a base class.

        This decorator does not add any functionality, but serves as documentation that communicates
        intentions behind how the class is intended to be used.
        """

        return func


# ----------------------------------------------------------------------
def EnsureValid(
    value: Optional[TypeT],
) -> TypeT:
    """Ensures that an optional value is not None and return it."""

    if value is None:
        raise ValueError("Invalid value")

    return value


# ----------------------------------------------------------------------
def EnsureValidList(
    value: Optional[list[TypeT]],
) -> list[TypeT]:
    """Ensures that a list is None or not empty and returns a list."""

    if value is None:
        return []

    if not value:
        raise ValueError("Invalid value")

    return value



# ----------------------------------------------------------------------
# |
# |  ExecuteTasks_UnitTest.py
# |
# |  David Brownell <db@DavidBrownell.com>
# |      2024-01-13 10:18:13
# |
# ----------------------------------------------------------------------
# |
# |  Copyright David Brownell 2024
# |  Distributed under the MIT License.
# |
# ----------------------------------------------------------------------
"""Unit tests for ExecuteTasks.py."""

# Testing Challenges
# ==================
# `ExecuteTasks.py` has proven to be very difficult to test for a couple of reasons:
#
# 1) The implementation executes across a number of threads which means the results are non-
#    deterministic.
#
# 2) The implementation relies on sys.stdout, the interactive nature of streams, and whether
#    a stream is headless or not. The way in which a test is run impacts all of these.
#
# Because of these factors, I have broken the tests into different categories:
#
# A) `ExecuteTasks_UnitTest.py`: Those things that can be verified automatically.
# B) `ExecuteTasks_TestManual.py`: Those things that cannot be verified automatically.
#
# Both of these tests should be run when introducing changes in `ExecuteTasks.py`.
#
# I wish that I knew of a better way to write these tests so that they didn't have to be run across
# multiple files and were fully automated. However, I have not been able to figure out a way to do
# that yet.
#
# To Run the Tests
# ================
#
# `ExecuteTasks_UnitTest.py`
# --------------------------
#
# Standard:
# `python Build.py Test`
#
# Code Coverage:
# `python Build.py Test --code-coverage`
#
# `ExecuteTasks_TestManual.py`
# ----------------------------
#
# Standard:
# `pytest tests/dbrownell_Common/ExecuteTasks_TestManual.py -vv --capture=no`
#
# Code Coverage:
# `pytest tests/dbrownell_Common/ExecuteTasks_TestManual.py -vv --capture=no --cov=dbrownell_Common.ExecuteTasks`
#
# Code Coverage with lcov output:
# `pytest tests/dbrownell_Common/ExecuteTasks_TestManual.py -vv --capture=no --cov=dbrownell_Common.ExecuteTasks --cov-report=lcov:tests\dbrownell_Common\lcov.info`
#

import re
import textwrap

from io import StringIO
from pathlib import Path
from typing import Any, Callable, cast

import pytest

from dbrownell_Common.ExecuteTasks import *
from dbrownell_Common import PathEx
from dbrownell_Common.Streams.DoneManager import DoneManager
from dbrownell_Common.Streams.TextWriter import TextWriter


# ----------------------------------------------------------------------
def test_ExecuteTasksSink():
    sink = _CreateSink()

    with DoneManager.Create(sink, "Executing tasks") as dm:
        results = _ExecuteTasks(dm, lambda x: (x * 2, 0))

    for index, result in enumerate(results):
        assert result == index * 2

    assert _Scrub(sink.getvalue()) == textwrap.dedent(
        """\
        Executing tasks...
          Tasks (5 items)...DONE! (0, <Scrubbed Time>, 5 items succeeded, no items with errors, no items with warnings)
        DONE! (0, <Scrubbed Time>)
        """,
    )


# ----------------------------------------------------------------------
def test_ExecuteTasksStream():
    stream = _FakeStream()

    with DoneManager.Create(stream, "Executing tasks") as dm:
        results = _ExecuteTasks(dm, lambda x: (x * 2, 0))

    for index, result in enumerate(results):
        assert result == index * 2

    stream.content[6] = _Scrub(cast(str, stream.content[6]))
    stream.content[-1] = _Scrub(cast(str, stream.content[-1]))

    assert stream.content == [
        "Executing tasks...",
        None,
        "\n",
        "  ",
        "Tasks (5 items)...",
        None,
        "DONE! (0, <Scrubbed Time>, 5 items succeeded, no items with errors, no items with warnings)",
        "\n",
        None,
        "DONE! (0, <Scrubbed Time>)\n",
    ]


# ----------------------------------------------------------------------
@pytest.mark.parametrize("no_compress_tasks", [False, True])
def test_TransformTasks(no_compress_tasks):
    sink = _CreateSink()

    num_tasks = int(multiprocessing.cpu_count() * 1.5)

    with DoneManager.Create(sink, "Transforming tasks") as dm:
        results = _TransformTasks(
            dm,
            lambda x, _: x * 2,
            num_tasks=num_tasks,
            no_compress_tasks=no_compress_tasks,
        )

    for index, result in enumerate(results):
        assert result == index * 2

    assert _Scrub(sink.getvalue()) == textwrap.dedent(
        """\
        Transforming tasks...
          Tasks ({num_tasks} items)...DONE! (0, <Scrubbed Time>, {num_tasks} items succeeded, no items with errors, no items with warnings)
        DONE! (0, <Scrubbed Time>)
        """,
    ).format(
        num_tasks=num_tasks,
    )


# ----------------------------------------------------------------------
def test_YieldQueueExecutor():
    sink = _CreateSink()

    results = [
        None,
    ] * 10

    with DoneManager.Create(sink, "Queue Executor") as dm:
        with YieldQueueExecutor(dm, "Queue") as executor:
            for index in range(len(results)):
                # ----------------------------------------------------------------------
                def Prepare(
                    on_simple_status_func: Callable[[str], None],
                    index=index,
                ) -> YieldQueueExecutorTypes.ExecuteFuncType:
                    # ----------------------------------------------------------------------
                    def Execute(status: Status) -> Optional[str]:
                        results[index] = index
                        return None

                    # ----------------------------------------------------------------------

                    return Execute

                # ----------------------------------------------------------------------

                executor(str(index), Prepare)

    for index, result in enumerate(results):
        assert result == index


# ----------------------------------------------------------------------
# |
# |  Private Types
# |
# ----------------------------------------------------------------------
class _FakeStream(TextWriter):
    # ----------------------------------------------------------------------
    def __init__(
        self,
        *,
        fileno: int = 123,
        isatty: bool = False,
        is_headless: bool = True,
        is_interactive: bool = False,
        supports_colors: bool = False,
    ):
        self.content: list[Optional[str]] = []
        self._fileno = fileno
        self._isatty = isatty

        Capabilities(
            stream=self,
            is_headless=is_headless,
            is_interactive=is_interactive,
            supports_colors=supports_colors,
        )

    # ----------------------------------------------------------------------
    def isatty(self) -> bool:
        return self._isatty

    # ----------------------------------------------------------------------
    def write(
        self,
        content: str,
    ) -> int:
        self.content.append(content)
        return len(content)

    # ----------------------------------------------------------------------
    def flush(self) -> None:
        self.content.append(None)

    # ----------------------------------------------------------------------
    def fileno(self) -> int:
        return self._fileno

    # ----------------------------------------------------------------------
    def close(self) -> None:
        pass


# ----------------------------------------------------------------------
# |
# |  Private Functions
# |
# ----------------------------------------------------------------------
def _ExecuteTasks(
    dm: DoneManager,
    execute_func: Callable[
        [int],  # context
        tuple[int, int],  # updated context, return code
    ],
    **execute_tasks_kwargs: Any,
) -> list[int]:
    task_data = [
        TaskData(
            str(value),
            value,
            None,
        )
        for value in range(5)
    ]

    results = [
        None,
    ] * len(task_data)

    # ----------------------------------------------------------------------
    def Init(context: Any) -> tuple[Path, ExecuteTasksTypes.PrepareFuncType]:
        # ----------------------------------------------------------------------
        def Prepare(on_simple_status_func: Callable[[str], None]) -> ExecuteTasksTypes.ExecuteFuncType:
            # ----------------------------------------------------------------------
            def Execute(status: Status) -> int:
                result, return_code = execute_func(context)

                results[context] = result
                return return_code

            # ----------------------------------------------------------------------

            return Execute

        # ----------------------------------------------------------------------

        return PathEx.CreateTempFileName(), Prepare

    # ----------------------------------------------------------------------

    ExecuteTasks(dm, "Tasks", task_data, Init, **execute_tasks_kwargs)

    return cast(list[int], results)


# ----------------------------------------------------------------------
def _TransformTasks(
    dm: DoneManager,
    transform_func: Callable[[int, Status], int],
    *,
    num_tasks: int = 5,
    **transform_tasks_kwargs: Any,
) -> list[int]:
    results = TransformTasks(
        dm,
        "Tasks",
        [TaskData(str(value), value) for value in range(num_tasks)],
        transform_func,
        **transform_tasks_kwargs,
    )

    return cast(list[int], results)


# ----------------------------------------------------------------------
def _CreateSink(
    supports_colors: bool = False,
) -> StringIO:
    sink = StringIO()

    Capabilities(
        stream=sink,
        is_headless=True,
        is_interactive=False,
        supports_colors=supports_colors,
    )

    return sink


# ----------------------------------------------------------------------
def _Scrub(
    content: str,
) -> str:
    return re.sub(
        r"\d+:\d{2}:\d{2}(?:\.\d+)?",
        "<Scrubbed Time>",
        content,
    )
